from trame.app import get_server
from trame.ui.html import DivLayout
from trame.widgets import html


def test_ui_vnode():
    server = get_server("test_ui_vnode")
    state = server.state
    state.ready()

    with DivLayout(server) as layout:
        server.ui.content(layout)

    assert state.trame__template_main == "<div >\n\n</div>"

    with server.ui.content:
        html.Div("Hello")

    assert state.trame__template_main == "<div >\n<div >\nHello\n</div>\n</div>"

    with server.ui.content:
        html.Div("World")

    assert (
        state.trame__template_main
        == "<div >\n<div >\nHello\n</div>\n<div >\nWorld\n</div>\n</div>"
    )

    with server.ui["content"].clear():
        html.Div("World")

    assert state.trame__template_main == "<div >\n<div >\nWorld\n</div>\n</div>"

    server.ui.clear()
    assert state.trame__template_main == "<div >\n<div >\nWorld\n</div>\n</div>"

    server.ui.flush_content()
    assert state.trame__template_main == "<div >\n\n</div>"

    server.ui.clear_layouts()



from argparse import ArgumentParser

from trame.app import get_server


def main():
    parser = ArgumentParser()
    parser.add_argument("--f", dest="foo")
    parser.parse_known_args()

    server = get_server("test_cli")
    server.start(timeout=1, open_browser=False, port=0)


if __name__ == "__main__":
    main()



import asyncio
import os
import subprocess
import sys
from pathlib import Path

import pytest
from trame.app import get_server
from trame.modules import www
from wslink import register as export_rpc
from wslink.websocket import LinkProtocol


@pytest.mark.asyncio
async def test_child_server():
    server = get_server("test_child_server")
    server.start(exec_mode="task", port=0)
    child_server = server.create_child_server(prefix="child_")

    assert await server.ready
    assert await child_server.ready

    assert server.running
    assert child_server.running

    server.state.a = 1
    child_server.state.a = 2

    assert server.state.has("a")
    assert server.state.has("child_a")
    assert child_server.state.has("a")

    assert server.state.child_a == child_server.state.a

    server.state.flush()
    await server.network_completion

    assert server.get_server_state() == {
        "name": "test_child_server",
        "state": {
            "a": 1,
            "child_a": 2,
            "trame__busy": 1,
            "trame__client_only": [
                "trame__busy",
            ],
            "trame__favicon": None,
            "trame__module_scripts": [],
            "trame__mousetrap": [],
            "trame__scripts": [],
            "trame__styles": [],
            "trame__title": "Trame",
            "trame__vue_use": [],
        },
    }

    await asyncio.sleep(0.1)
    await server.stop()


def test_http_headers():
    server = get_server("test_http_headers")

    server.http_headers.shared_array_buffer = True
    server.http_headers.set_header("hello", "world")
    server.http_headers.set_header("hello2", "world2")
    server.http_headers.remove_header("hello2")

    assert server.http_headers.headers == {
        "hello": "world",
        "Cross-Origin-Opener-Policy": "same-origin",
        "Cross-Origin-Embedder-Policy": "require-corp",
        "Access-Control-Allow-Origin": "*",
    }
    server.http_headers.shared_array_buffer = False
    assert server.http_headers.headers == {
        "hello": "world",
    }
    assert server.http_headers.get_header("hello") == "world"


def test_enable_module():
    server = get_server("test_enable_module")
    child_server = server.create_child_server(prefix="child_")

    module = {
        "scripts": ["fake_url/script.js"],
        "state": {
            "a": 1,
            "b": 2,
        },
        "serve": {"data": "/tmp"},
    }

    assert child_server.enable_module(module)
    assert child_server.enable_module(www)

    # should skip since already loaded
    assert not server.enable_module(module)
    assert not server.enable_module(www)

    assert server.state.a == 1
    assert server.state.b == 2
    serve_conf = {
        k: v
        for k, v in server.serve.items()
        if not k.startswith("__trame_client_external_scripts")
    }
    assert serve_conf == {"data": "/tmp"}

    @server.change("a")
    def on_change(**_):
        pass

    @server.trigger("my_name")
    def another_method():
        pass

    assert server.state._change_callbacks["a"][0][0] == on_change
    assert server.trigger_name(another_method) == "my_name"
    assert server.name == "test_enable_module"

    # default is vue3
    assert server.client_type == "vue3"

    # can still be overridden
    server.client_type = "vue2"
    assert server.client_type == "vue2"

    # Can only be set once
    with pytest.raises(TypeError):
        server.client_type = "vue3"


def test_cli_args_collision(pytestconfig: pytest.Config):
    cli = pytestconfig.rootpath / "tests/data/test_cli.py"
    subprocess.run([sys.executable, str(cli), "--f", "foo"], check=True)


def test_cli():
    server = get_server("test_cli")
    child_server = server.create_child_server(prefix="child_")
    server.cli.add_argument("--data")
    child_server.cli.add_argument("--data2")
    args = server.cli.parse_known_args()[0]
    assert args.data is None
    assert args.data2 is None


@pytest.mark.asyncio
async def test_server_start_async():
    server = get_server("test_server_start_async")
    count = 0

    def on_start(s):
        nonlocal count
        count += 2
        assert server is s

    def on_ready(**_):
        nonlocal count
        count += 3

    child_server = server.create_child_server(prefix="child_")

    server.controller.on_server_start.add(on_start)
    server.controller.on_server_ready.add(on_ready)

    class TestProto(LinkProtocol):
        @export_rpc("pytest.protocol.test")
        def run(
            self,
        ):
            return 11

    def register_protocol(protocol):
        protocol.registerLinkProtocol(TestProto())

    child_server.add_protocol_to_configure(register_protocol)

    server.state.a = 10

    child_server.start(exec_mode="task", thread=True, port=0)

    assert await server.ready
    assert await child_server.ready

    # Should be a noop as already started
    server.start(exec_mode="task", port=0)

    assert server.protocol_call("pytest.protocol.test") == 11

    await asyncio.sleep(0.1)
    assert count == 5

    server.force_state_push("a")
    server.js_call("js_ref", "method", "arg1", "arg2")

    server.clear_state_client_cache("a")

    assert child_server.protocol == server.protocol
    assert child_server.port == server.port
    assert child_server.port != 0

    await child_server.stop()


def test_server_start_sync():
    os.environ["TRAME_ARGS"] = "--banner --no-http"
    os.environ["TRAME_LOG_NETWORK"] = "trame_net.log"
    server = get_server("test_server_start_sync")
    server.serve.update(
        {
            "data": str(Path(__file__).parent.resolve()),
            "data2": (
                str(Path(__file__).parent.resolve()),
                "sync",
            ),  # don't remember usage...
        }
    )
    server.state.a = b"sdkfjhvlskdjhf"
    server.start(timeout=1, open_browser=False, port=0)


def test_ui():
    server = get_server("test_ui")
    server.ui.vnode  # noqa: B018


@pytest.mark.asyncio
async def test_server_ready_forwards_exceptions_in_ready_future():
    server_1 = get_server("test_ready_exception_forwarding_1")
    server_1.start(exec_mode="task", port=0)
    await server_1.ready

    server_2 = get_server("test_ready_exception_forwarding_2")
    server_2.start(exec_mode="task", port=server_1.port)
    with pytest.raises(OSError, match="error while attempting to bind on address"):
        await server_2.ready


@pytest.mark.asyncio
async def test_server_task_cancel_doesnt_hang_ready_future():
    server = get_server("test_cancel")
    task = server.start(exec_mode="task", port=0)
    task.cancel()
    await asyncio.wait_for(server.ready, timeout=1)



from __future__ import annotations  # triggers lazy evaluation of field.type

from dataclasses import dataclass
from enum import Enum, auto

import pytest

from trame_server import Server
from trame_server.utils.typed_state import TypedState


class AnEnum(Enum):
    A = auto()
    B = auto()


@dataclass
class DataWithTypesAnnotations:
    my_enum: AnEnum


@pytest.fixture
def state():
    server = Server()
    server.state.ready()
    return server.state


def test_is_compatible_with_from_future_annotations(state):
    typed_state = TypedState(state, DataWithTypesAnnotations)
    typed_state.data.my_enum = AnEnum.B
    assert typed_state.data.my_enum == AnEnum.B
    assert state[typed_state.name.my_enum] == AnEnum.B.value



import logging
import sys
import traceback

from trame.app import TrameApp
from trame.decorators import life_cycle
from trame.ui.html import DivLayout
from trame.widgets import html

# Disable default stdout exception
logging.getLogger("wslink.protocol").setLevel(logging.CRITICAL)


class MonitorException(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)
        self._build_ui()

    def _build_ui(self):
        with DivLayout(self.server) as self.ui:
            html.Button("Divide by 0", click=self.divide_by_zero)
            html.Button(
                "No method",
                click="trame.client.getConnection().getSession().call('no_rpc_endpoint', [], {})",
            )

    @life_cycle.server_ready
    def _setup_wslink_listener(self, **_):
        self.server.protocol.log_emitter.add_event_listener(
            "exception", self.on_exception
        )
        self.server.protocol.log_emitter.add_event_listener("error", self.on_error)

    def divide_by_zero(self):
        return 1 / 0

    def on_exception(self, exception):
        print(">" * 60)
        print("exception ->", exception)
        traceback.print_exception(
            type(exception), exception, exception.__traceback__, file=sys.stdout
        )
        print("<" * 60)

    def on_error(self, msg):
        print(">" * 60)
        print("error ->", msg)
        print("<" * 60)


def main():
    app = MonitorException()
    app.server.start()


if __name__ == "__main__":
    main()



from trame.app import get_server
from trame.decorators import TrameApp, change
from trame.ui.vuetify3 import SinglePageLayout
from trame.widgets import client
from trame.widgets import vuetify3 as v3

import trame_server


@TrameApp()
class MainApp:
    def __init__(self, server=None):
        self.server = get_server(server)
        self._build_ui()

        SecondApp(self.server)

    def _build_ui(self):
        with SinglePageLayout(self.server, full_height=True) as layout:
            with layout.content:
                with client.SizeObserver("main_size_observer"):
                    with v3.VContainer():
                        v3.VBtn(
                            "Open Second App",
                            click="window.open('/?ui=second', target='_blank')",
                        )

    @change("main_size_observer")
    def main_size_observer(self, main_size_observer, **_):
        print("main_size_observer", main_size_observer)


@TrameApp()
class SecondApp:
    def __init__(
        self, server: trame_server.Server | None = None, template_name="second"
    ):
        self.prefix = ""
        if server:
            self.prefix = "second_"
            self.server = server.create_child_server(prefix=self.prefix)
        else:
            self.server = get_server(server)

        self.state = self.server.state
        self.ctrl = self.server.controller

        self._build_ui(template_name)
        # server.state.change("second_size_observer")(lambda **_: print("second size"))

    @change("second_size_observer")
    def second_size_observer(self, **_):
        print("second_size_observer", self.state.second_size_observer)

    def _build_ui(self, template_name):
        # self.state.dialog_show = False
        with SinglePageLayout(
            self.server, template_name=template_name, full_height=True
        ) as layout:
            with layout.content:
                with client.SizeObserver("second_size_observer"):
                    with v3.VContainer():
                        v3.VBtn("Hello")


def main():
    app = MainApp()
    app.server.start()


if __name__ == "__main__":
    main()



import asyncio
import io
from contextlib import redirect_stdout

import pytest
from trame.app import get_server

from trame_server import utils
from trame_server.utils import banner, hot_reload, server, version


def test_banner():
    with io.StringIO() as buf, redirect_stdout(buf):
        banner.print_banner()
        output = buf.getvalue()
        assert len(output) > 4096


@pytest.mark.asyncio
async def test_print_informations():
    server_app = get_server("test_print_informations")

    with io.StringIO() as buf, redirect_stdout(buf):
        server_app.start(exec_mode="task", port=0)
        assert await server_app.ready

        # not automatic when server start as task
        server.print_informations(server_app)
        output = buf.getvalue()
        assert "http://localhost:" in output

        await asyncio.sleep(0.1)
        await server_app.stop()


def test_version():
    from trame_client import __version__ as v_client  # noqa: PLC0415

    from trame_server import __version__ as v_server  # noqa: PLC0415

    assert version.get_version("trame_server") == v_server
    assert version.get_version("trame_client") == v_client
    assert version.get_version("trame").split(".")[0] == "3"
    assert version.get_version("something_that_does_not_exist") is None


def test_utils_fn():
    assert utils.isascii(b"sdv")
    assert utils.isascii("sdv")
    assert not utils.isascii("Hällö, Wörld!")

    # remove filter keys
    input_dict = {"a": "hello", "b": "world", "_filter": ["b"]}
    output_dict = utils.clean_value(input_dict)
    assert input_dict != output_dict
    assert output_dict == {"a": "hello", "_filter": ["b"]}

    # deep clone
    input_dict = {
        "a": {
            "a.a": {"x": 1},
            "a.b": {"y": 2},
        },
        "b": {
            "b.a": {"z": 3},
            "b.b": {"xyz": 4},
        },
    }
    output_dict = {}
    utils.update_dict(output_dict, input_dict)
    assert output_dict == input_dict
    assert input_dict["a"] is input_dict["a"]
    assert output_dict["a"] == input_dict["a"]
    assert output_dict["a"] is not input_dict["a"]
    assert output_dict["a"]["a.a"] == input_dict["a"]["a.a"]
    assert output_dict["a"]["a.a"] is not input_dict["a"]["a.a"]

    # reduce vue_use
    class FakeState:
        def __init__(self):
            self.trame__vue_use = [
                "trame_vtk",
                ("hello", {"a": 1, "c": {"x": 1}}),
                "trame_vtk",
                ("hello", {"b": 2, "c": {"y": 2}}),
                "trame_vtk",
                "trame_xyz",
            ]

    fake_state = FakeState()
    utils.reduce_vue_use(fake_state)
    assert fake_state.trame__vue_use == [
        "trame_vtk",
        (
            "hello",
            {
                "a": 1,
                "b": 2,
                "c": {
                    "x": 1,
                    "y": 2,
                },
            },
        ),
        "trame_xyz",
    ]


def test_hot_reload():
    @hot_reload.hot_reload
    def re_eval():
        pass

    # skip decorate twice
    hot_reload.hot_reload(re_eval)

    re_eval()



"""
Usage example of the trame_server.utils.typed_state.TypedState class.
This class provides a convenient way to create two-way bindings between a trame state and a dataclass.
"""

from dataclasses import dataclass, field
from datetime import datetime, time
from enum import Enum, auto
from uuid import UUID, uuid4

from trame.app import get_server
from trame.ui.vuetify3 import SinglePageLayout
from trame.widgets import vuetify3 as vuetify

from trame_server.utils.typed_state import (
    DefaultEncoderDecoder,
    TypedState,
)

server = get_server(client_type="vue3")
state, ctrl = server.state, server.controller


class Priority(Enum):
    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()


def current_time() -> time:
    return datetime.now().time()


@dataclass
class Inner:
    """
    Dataclasses can define default values.
    The default values will be available in the state as default values.
    """

    text_value: str = ""
    priority: Priority = Priority.MEDIUM
    tags: list[str] = field(default_factory=list)
    selected_time: time = field(default_factory=current_time)


@dataclass
class TrameState:
    """
    Dataclasses can be nested and support various types, including enums and UUIDs.
    """

    inner: Inner = field(default_factory=Inner)
    last_inner_update: datetime = field(default_factory=datetime.now)
    task_id: UUID = field(default_factory=uuid4)


class CustomEncoder(DefaultEncoderDecoder):
    def encode(self, obj):
        if isinstance(obj, time):
            return obj.isoformat(timespec="minutes")
        return super().encode(obj)


# Create TypedState instance with custom encoder
typed_state = TypedState(state, TrameState, encoders=[CustomEncoder()])


# Nested names will contain the nested field names automatically.
print(f"State field path: {typed_state.name.inner.text_value}")


# Nested instances can be split into separate typed states if needed
inner_state = typed_state.get_sub_state(typed_state.name.inner)
assert isinstance(inner_state, TypedState)
assert inner_state.name.text_value == typed_state.name.inner.text_value


# State change callbacks
def on_text_change(text_value):
    print(f"Current text: {text_value}")


def on_priority_change(priority):
    print(f"Priority changed to: {priority}")


def on_task_update(text_value: str, priority: Priority, tags: list[str]):
    print(f"Task updated - Text: {text_value}, Priority: {priority}, Tags: {tags}")


def on_inner_data_change(inner_data: Inner):
    # When binding nested dataclass, the callback uses typed_state.data.inner
    assert inner_data == typed_state.data.inner

    # Nested dataclasses can be converted back to a dataclass using the as_dataclass method
    print(f"Inner data changed: {TypedState.as_dataclass(inner_data)}")

    typed_state.data.last_inner_update = datetime.now()


def on_time_change(last_modified_time: time):
    print(f"Selected time changed: {last_modified_time}")


# TypedState provides a bind_changes method to bind state changes to strongly typed callbacks
# Notice that the key binding is not limited to dataclass leaf but includes nested classes providing flexibility to
# the binding approach.
typed_state.bind_changes(
    {
        typed_state.name.inner.text_value: on_text_change,
        typed_state.name.inner.priority: on_priority_change,
        typed_state.name.inner.selected_time: on_time_change,
        (
            typed_state.name.inner.text_value,
            typed_state.name.inner.priority,
            typed_state.name.inner.tags,
        ): on_task_update,
        (typed_state.name.inner,): on_inner_data_change,
    }
)


@ctrl.add("reset_task")
def reset_task():
    # Dataclass values can be set from a dataclass instance using the TypedState.from_dataclass method
    TypedState.from_dataclass(typed_state.data.inner, Inner())
    typed_state.data.task_id = uuid4()


with SinglePageLayout(server) as layout:
    with layout.content:
        with vuetify.VContainer(
            fluid=True,
            classes="d-flex justify-center align-center",
            style="height: 100%;",
        ):
            with vuetify.VCard(style="max-width: 500px;"):
                vuetify.VCardTitle("TypedState trame example")
                with vuetify.VCardText():
                    # Task description input
                    vuetify.VTextField(
                        v_model=(typed_state.name.inner.text_value,),
                        label="Task Description",
                        outlined=True,
                    )

                    # Priority selector
                    # Typed state encoder can be used to encode data in a consistent way for reading in the callbacks.
                    # Here the VSelect PRIORITY Enum value will be encoded by the typed_state encoder.
                    vuetify.VSelect(
                        v_model=(typed_state.name.inner.priority,),
                        items=(
                            "options",
                            typed_state.encode(
                                [{"text": p.name.title(), "value": p} for p in Priority]
                            ),
                        ),
                        item_title="text",
                        item_value="value",
                        label="Priority",
                        outlined=True,
                    )

                    # Tags input
                    vuetify.VCombobox(
                        v_model=(typed_state.name.inner.tags,),
                        label="Tags",
                        multiple=True,
                        chips=True,
                        clearable=True,
                        outlined=True,
                        style="width: 100%; max-width: 100%;",
                    )

                    # Time picker
                    vuetify.VTextField(
                        v_model=(typed_state.name.inner.selected_time,),
                        label="Selected Time",
                        type="time",
                        outlined=True,
                    )

                    # Display current task ID
                    vuetify.VTextField(
                        v_model=(typed_state.name.task_id,),
                        label="Task ID",
                        readonly=True,
                        outlined=True,
                    )

                    # Display last modification date
                    vuetify.VTextField(
                        v_model=(typed_state.name.last_inner_update,),
                        label="Last Updated",
                        readonly=True,
                        outlined=True,
                    )

                with vuetify.VCardActions():
                    vuetify.VBtn(
                        "Reset Task",
                        color="primary",
                        click=ctrl.reset_task,
                    )


server.start()



import time
from pathlib import Path

from trame.app import get_server
from trame.ui.vuetify import SinglePageLayout
from trame.widgets import vuetify

server = get_server(client_type="vue2")
state, ctrl = server.state, server.controller

state.trame_title = "Server"
state.counter = 0
state.attch_data = None


@state.change("test_file")
def on_file(test_file, **_):
    if test_file is not None:
        print("Got file...", test_file.get("name"))
    else:
        print("No file...")


@state.change("counter")
def on_change(counter, **_):
    msg = f"server::counter = {counter}"
    if state.log is None:
        state.log = ""
    state.log += msg + "\n"
    print(msg)


@ctrl.trigger("my_method")
def on_method(*args, **kwargs):
    msg = f"server::method {args} {kwargs}"
    if state.log is None:
        state.log = ""
    state.log += msg + "\n"
    print(msg)
    return f"It is {time.time()}s"


def test_attachment():
    content = Path(__file__)
    state.attch_data = {
        "id": state.counter,
        "time": time.time(),
        "big": server.protocol.addAttachment(content.read_bytes()),
    }


with SinglePageLayout(server) as layout:
    layout.title.set_text(state.trame_title)

    with layout.toolbar as toolbar:
        vuetify.VSpacer()
        toolbar.add_child("{{ counter }}")
        vuetify.VSpacer()
        vuetify.VBtn("Test attachment", click=test_attachment)
        vuetify.VFileInput(
            v_model=("test_file", None),
            dense=True,
            hide_details=True,
            style="max-width: 300px;",
            classes="mx-2",
        )
        vuetify.VBtn("-", click="counter--")
        vuetify.VBtn("+", click="counter++")
        vuetify.VBtn("fn", click="trigger('my_method', [counter, 2], {a: 1, b: 2})")

    with layout.content:
        with vuetify.VContainer(fluid=True):
            vuetify.VTextarea(
                v_model=("log", ""),
                outlined=True,
                clearable=True,
                auto_grow=True,
            )

if __name__ == "__main__":
    server.start()



import asyncio
import weakref
from unittest.mock import MagicMock

import pytest

from trame_server.core import Translator
from trame_server.state import State


class FakeServer:
    def __init__(self):
        self._change_callbacks = {}
        self._events = []
        self.translator = Translator()

    def _push_state(self, delta_state):
        self._events.append({"type": "push", "content": {**delta_state}})

    def add_event(self, content, type="msg"):
        self._events.append({"type": type, "content": content})

    @property
    def pushed_state(self) -> dict:
        pushed = {}
        for e in self._events:
            if e["type"] == "push":
                pushed.update(e["content"])
        return pushed

    def __repr__(self) -> str:
        lines = [""]
        for line_nb, entry in enumerate(self._events):
            lines.append(f"{line_nb:6} {entry.get('type'):5}: {entry.get('content')}")
        lines.append("")
        return "\n".join(lines)


@pytest.fixture
def server():
    return FakeServer()


@pytest.fixture
def state(server):
    _state = State(commit_fn=server._push_state)
    _state.ready()
    return _state


def test_minimum_change_detection():
    """
     0 msg  : test_minimum_change_detection
     1 msg  : Before server ready
     2 push : {'a': 2}
     3 exec : 2
     4 msg  : After server ready
     5 msg  : (prev=2) After 2, 3, 3, 4, 4
     6 msg  : (prev=4) Before Flush
     7 push : {'a': 4}
     8 exec : 4
     9 msg  : (prev=4) After Flush
    10 msg  : Enter with a=4
    11 msg  : About to exit a=4
    12 msg  : (prev=4) After with state + same value
    13 msg  : Enter with a=4
    14 msg  : About to exit a=5
    15 push : {'a': 5}
    16 exec : 5
    17 msg  : (prev=5) After with state + 3,4,5
    18 msg  : Enter with a=5
    19 msg  : About to exit a=5
    20 msg  : (prev=5) After with state + 3,5,4,5
    21 msg  : Enter with a=5
    22 msg  : About to exit a=5
    23 push : {'b': 3, 'c': 2}
    24 msg  : (prev=5) After with state + a:1,5 b:2,3 c:3,2
    25 msg  : Enter with a=5
    26 msg  : About to exit a=1
    27 push : {'a': 1, 'b': 2, 'c': 3}
    28 exec : 1
    """
    server = FakeServer()
    server.add_event("test_minimum_change_detection")
    state = State(commit_fn=server._push_state)

    @state.change("a")
    def on_change_exec(a, **_):
        server.add_event(type="exec", content=a)

    state.a = 1
    state.a = 1
    state.a = 2
    state.a = 2

    server.add_event("Before server ready")
    state.ready()
    server.add_event("After server ready")

    state.a = 2
    state.a = 3
    state.a = 3
    state.a = 4
    state.a = 4

    server.add_event("(prev=2) After 2, 3, 3, 4, 4")

    # Flush
    server.add_event("(prev=4) Before Flush")
    state.flush()
    server.add_event("(prev=4) After Flush")

    # This should be a NoOp
    with state:
        server.add_event(f"Enter with a={state.a}")
        state.a = 4
        server.add_event(f"About to exit a={state.a}")

    server.add_event("(prev=4) After with state + same value")

    with state:
        server.add_event(f"Enter with a={state.a}")
        state.a = 3
        state.a = 4
        state.a = 5
        server.add_event(f"About to exit a={state.a}")

    server.add_event("(prev=5) After with state + 3,4,5")

    # Even though it changed, finally it is the same value
    with state:
        server.add_event(f"Enter with a={state.a}")
        state.a = 3
        state.a = 5
        state.a = 4
        state.a = 5
        server.add_event(f"About to exit a={state.a}")

    server.add_event("(prev=5) After with state + 3,5,4,5")

    # Use update to set {a: 1, b: 2, c: 3}
    with state:
        server.add_event(f"Enter with a={state.a}")
        state.update({"a": 1, "b": 2, "c": 3})
        state.update({"a": 5, "b": 3, "c": 2})
        server.add_event(f"About to exit a={state.a}")

    server.add_event("(prev=5) After with state + a:1,5 b:2,3 c:3,2")

    # Use update to set {a: 1, b: 2, c: 3}
    with state:
        server.add_event(f"Enter with a={state.a}")
        state.update({"a": 1, "b": 2, "c": 3})
        server.add_event(f"About to exit a={state.a}")

    # Validate event
    result = [line.strip() for line in str(server).split("\n")]
    expected = [
        line.strip() for line in str(test_minimum_change_detection.__doc__).split("\n")
    ]

    # Grab new scenario output
    # print(expected)
    # print("-"*60)
    # print(result)

    assert expected == result


def test_client_only():
    server = FakeServer()
    server.add_event("test_client_only")
    state = State(commit_fn=server._push_state)
    state.ready()

    state.aa = 1
    state.client_only("aa")


def test_dict_api():
    server = FakeServer()
    server.add_event("test_dict_api")
    state = State(commit_fn=server._push_state)
    state.flush()  # should return right away since not ready
    state.ready()

    state.a = 1
    state.c = []
    assert state.has("a")
    assert not state.has("b")
    state.setdefault("a", 10)
    state.setdefault("b", 20)
    assert state.has("b")
    assert state.a == 1
    assert state.b == 20

    assert state.is_dirty_all("a", "b")
    assert state.is_dirty("a", "b")
    state.flush()
    assert not state.is_dirty("a", "b")
    assert state.setdefault("a", 30) == 1

    state.c.append("item")
    assert not state.is_dirty("c")
    state.dirty("c")
    assert state.is_dirty("c")

    assert state.initial == {"a": 1, "b": 20, "c": ["item"]}


@pytest.mark.asyncio
async def test_change_detection():
    """
    0 msg  : test_change_detection
    1 push : {'a': 2}
    2 msg  : a changed (sync)
    3 msg  : a changed (async)
    """
    server = FakeServer()
    server.add_event("test_change_detection")
    state = State(commit_fn=server._push_state, hot_reload=True)
    state.ready()

    state.a = 1

    @state.change("a")
    def regular_callback(**__kwargs):
        server.add_event("a changed (sync)")

    @state.change("a")
    async def coroutine_callback(**__kwargs):
        server.add_event("a changed (async)")

    assert "a" in state._pending_update
    state.clean("a")
    assert "a" not in state._pending_update

    with state:
        state.a = 2

    await asyncio.sleep(0.1)

    result = [line.strip() for line in str(server).split("\n")]
    expected = [line.strip() for line in str(test_change_detection.__doc__).split("\n")]

    # Grab new scenario output
    # print(expected)
    # print("-"*60)
    # print(result)

    assert expected == result


def test_dunder():
    server = FakeServer()
    server.add_event("test_dunder")
    state = State(commit_fn=server._push_state, hot_reload=True)
    state.ready()

    # get dunder
    assert state.__dict__ != state.__getattr__("__dict__")

    # get private (not in state)
    assert state._something is None

    # set private (not in state)
    state._something = 1
    assert state._something == 1

    state.flush()
    assert state.to_dict() == {}


@pytest.mark.asyncio
async def test_modified_keys():
    """
     0 msg  : test_modified_keys
     1 push : {'a': 1, 'b': 2, 'c': 3}
     2 msg  : get initial a,b,c
     3 msg  : changed should be => a
     4 push : {'a': 2}
     5 msg  : changed ['a']
     6 msg  : End of flush 1
     7 msg  : changed should be => a, b
     8 push : {'a': 3, 'b': 4}
     9 msg  : changed ['a', 'b']
    10 msg  : End of flush 2
    11 msg  : changed should be => a, b, c
    12 push : {'a': 4, 'b': 6, 'c': 6}
    13 msg  : changed ['a', 'b', 'c']
    14 msg  : side effect c => a + b
    15 push : {'a': 4.5, 'b': 6.5}
    16 msg  : changed ['a', 'b']
    17 msg  : End of flush 3
    """
    server = FakeServer()
    server.add_event("test_modified_keys")
    state = State(commit_fn=server._push_state)

    NAMES = ["a", "b", "c"]
    state.update(
        {
            "a": 1,
            "b": 2,
            "c": 3,
        }
    )
    state.ready()
    server.add_event("get initial a,b,c")
    await asyncio.sleep(0.01)

    @state.change(*NAMES)
    def on_change(**_):
        m_keys = list(state.modified_keys)
        m_keys.sort()
        server.add_event(f"changed {m_keys}")

    @state.change("c")
    def trigger_side_effect(**_):
        server.add_event("side effect c => a + b")
        state.a += 0.5
        state.b += 0.5

    with state:
        state.a += 1
        server.add_event("changed should be => a")

    # yield
    await asyncio.sleep(0.01)
    server.add_event("End of flush 1")

    with state:
        state.a += 1
        state.b += 2
        server.add_event("changed should be => a, b")

    # yield
    await asyncio.sleep(0.01)
    server.add_event("End of flush 2")

    with state:
        state.a += 1
        state.b += 2
        state.c += 3
        server.add_event("changed should be => a, b, c")

    # yield
    await asyncio.sleep(0.1)
    server.add_event("End of flush 3")

    result = [line.strip() for line in str(server).split("\n")]
    expected = [line.strip() for line in str(test_modified_keys.__doc__).split("\n")]

    # sometime 13 and 14 could have a reverse execution order
    # as trame does not guaranty the execution order of the callbacks.
    result.pop(14)
    result.pop(14)
    expected.pop(14)
    expected.pop(14)

    print(result)

    assert expected == result


def test_weakref():
    server = FakeServer()
    state = State(commit_fn=server._push_state, hot_reload=True)
    state.ready()

    class Obj:
        method_call_count = 0
        destructor_call_count = 0

        def __del__(self):
            Obj.destructor_call_count += 1

        def fn(self, *_args, **_kwargs):
            Obj.method_call_count += 1
            print("Obj.fn called")
            return 1

    o = Obj()

    state.a = 1

    state.change("a")(weakref.WeakMethod(o.fn))

    state.a = 2
    state.flush()
    assert Obj.method_call_count == 1

    del o
    assert Obj.destructor_call_count == 1

    state.a = 3
    state.flush()
    assert Obj.method_call_count == 1


def test_given_explicit_keys_when_modified_inside_context_then_listeners_are_not_called(
    state, server
):
    mock = MagicMock()

    @state.change("a")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners("a"):
        state.a = 2

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["a"] == 2


def test_given_explicit_keys_when_flushing_inside_context_then_listeners_are_not_called(
    state, server
):
    mock = MagicMock()

    @state.change("a")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners("a"):
        for i in range(3):
            state.a = i
            state.flush()

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["a"] == 2


def test_given_no_keys_when_any_key_modified_inside_context_then_listeners_are_not_called(
    state, server
):
    mock = MagicMock()

    @state.change("a", "b")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners():
        state.a = 2
        state.b = 3

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["a"] == 2
    assert server.pushed_state["b"] == 3


def test_given_nested_no_keys_when_any_key_modified_inside_context_then_listeners_are_not_called(
    state, server
):
    mock = MagicMock()

    @state.change("a", "b")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners("d"):
        with state.suppress_change_listeners():
            with state.suppress_change_listeners("e"):
                state.a = 2
                state.b = 3

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["a"] == 2
    assert server.pushed_state["b"] == 3


def test_given_modified_keys_when_context_exits_then_keys_are_marked_clean_in_state(
    state,
):
    with state.suppress_change_listeners("a"):
        state.a = 2

    state.flush()
    assert not state.is_dirty("a")


def test_given_multiple_nested_suppress_contexts_when_exiting_then_all_keys_are_synced_correctly(
    state, server
):
    mock = MagicMock()

    @state.change("a", "b")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners("a"):
        state.a = 2
        state.flush()

        with state.suppress_change_listeners("b"):
            state.b = 3
            state.flush()

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["a"] == 2
    assert server.pushed_state["b"] == 3


def test_given_nested_suppress_with_same_key_keeps_suppression_in_outer_context(
    state, server
):
    mock = MagicMock()

    @state.change("a")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners("a"):
        with state.suppress_change_listeners("a"):
            state.a = 3
            state.flush()

        state.a = 2
        state.flush()

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["a"] == 2


def test_given_mixed_updates_when_some_are_suppressed_then_only_normal_updates_trigger_listeners(
    state,
):
    mock_a = MagicMock()
    mock_b = MagicMock()

    @state.change("a")
    def on_a(**_):
        mock_a()

    @state.change("b")
    def on_b(b, **_):
        mock_b(b)

    with state.suppress_change_listeners("a"):
        state.a = 2
        state.b = 3

    state.flush()
    mock_a.assert_not_called()
    mock_b.assert_called_once_with(3)


def test_given_exception_in_context_when_raised_then_cleanup_logic_still_executes(
    state, server
):
    mock = MagicMock()

    @state.change("a")
    def on_change(a, **_):
        mock(a)

    try:
        with state.suppress_change_listeners("a"):
            raise ValueError()
    except ValueError:
        pass

    state.a = 2
    state.flush()
    assert server.pushed_state["a"] == 2
    mock.assert_called_once_with(2)


def test_given_namespaced_state_when_using_suppress_context_then_keys_are_correctly_translated(
    server,
):
    translator = Translator(prefix="test_")
    state = State(translator=translator, commit_fn=server._push_state)
    state.ready()

    mock = MagicMock()

    @state.change("a")
    def on_change(**_):
        mock()

    with state.suppress_change_listeners("a"):
        state.a = 2

    state.flush()
    mock.assert_not_called()
    assert server.pushed_state["test_a"] == 2


def test_given_change_callback_when_it_modifies_state_suppressed_then_recursion_is_avoided(
    state, server
):
    mock = MagicMock()

    @state.change("a")
    def on_a(a, **_):
        with state.suppress_change_listeners("b"):
            state.b = a
            state.flush()

    @state.change("b")
    def on_b(**_):
        mock()

    state.a = 2
    state.flush()

    assert not state.is_dirty("b")
    assert server.pushed_state["b"] == 2
    mock.assert_not_called()


def test_given_chain_supress_sequence_correctly_suppress_scoped_listeners(state):
    mock = MagicMock()

    @state.change("a", "b", "c")
    def on_a(**kwargs):
        mock(**kwargs)

    with state.suppress_change_listeners():
        state.a = 1
        state.b = 2

    with state.suppress_change_listeners("c"):
        state.a = 42
        state.c = 3

    state.flush()
    mock.assert_called_once_with(a=42, b=2, c=3)


def test_keys_marked_dirty_triggers_listener_changes(state):
    mock = MagicMock()

    @state.change("a")
    def on_change(a, **_):
        mock(a)

    state.a = 1
    state.flush()
    mock.reset_mock()

    state.dirty("a")
    state.flush()
    mock.assert_called_once_with(1)


def test_keys_marked_dirty_in_suppress_does_not_trigger_listener_changes(state):
    mock = MagicMock()

    @state.change("a")
    def on_change(a, **_):
        mock(a)

    state.a = 1
    state.flush()
    mock.reset_mock()

    with state.suppress_change_listeners("a"):
        state.dirty("a")

    state.flush()
    mock.assert_not_called()


def test_keys_marked_clean_does_not_trigger_listener_changes(state):
    mock = MagicMock()

    @state.change("a")
    def on_change(a, **_):
        mock(a)

    state.a = 2
    state.clean("a")
    state.flush()
    mock.assert_not_called()


def test_keys_marked_clean_in_suppress_does_not_trigger_listener_changes(state):
    mock = MagicMock()

    @state.change("a")
    def on_change(a, **_):
        mock(a)

    state.a = 2
    with state.suppress_change_listeners("a"):
        state.clean("a")

    state.flush()
    mock.assert_not_called()


def test_state_change_listeners_are_triggered_in_modified_order(state):
    keys = [f"k_{i}" for i in range(10)]
    recorded = []

    def create_listener(k):
        def on_change(**_):
            recorded.append(k)

        return on_change

    for key in keys:
        state.change(key)(create_listener(key))

    for i_trial in range(1000):
        recorded.clear()

        for key in keys:
            state[key] = i_trial

        state.flush()

        _assert_msg = (
            f"Order mismatch at iteration {i_trial}: expected {keys}, got {recorded}"
        )
        assert recorded == keys, _assert_msg


def test_setdefault_trigger_change_on_ready():
    state = State()
    mock = MagicMock()

    @state.change("a")
    def on_change(a, **_):
        mock(a)

    state.setdefault("a", 1)
    mock.assert_not_called()

    state.ready()
    mock.assert_called_once_with(1)


def test_child_state_changes_propagated_to_parent_state(state):
    child_state = State(internal=state)

    mock1 = MagicMock()
    mock2 = MagicMock()

    @state.change("a")
    def on_state_change(**_):
        mock1()

    @child_state.change("a")
    def on_child_state_change(**_):
        mock2()

    child_state.a = 1
    state.flush()

    mock1.assert_called_once()
    mock2.assert_called_once()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "trame",
#     "trame-vtk",
#     "trame-vuetify",
# ]
# ///
from trame.app import TrameApp, demo
from trame.ui.html import DivLayout
from trame.widgets import html


class App(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)
        self.sub_app = demo.Cone("internal")
        self.task = None
        self._build_ui()

    async def start(self):
        if self.state.running:
            return

        self.sub_app.server.start(exec_mode="task", port=0)
        await self.sub_app.server.ready
        self.state.running = True
        port = self.sub_app.server.port
        self.state.url = f"http://localhost:{port}/"

    async def stop(self):
        if not self.state.running:
            return

        await self.sub_app.server.stop()
        self.state.running = False
        self.state.url = ""

    def _build_ui(self):
        with DivLayout(self.server) as self.ui:
            html.Button("Stop", click=self.stop, v_if=("running", False))
            html.Button("Start", click=self.start, v_else=True)
            html.A("{{ url }}", href=("url", ""), v_show="running", target="_blank")


def main():
    app = App()
    app.server.start()


if __name__ == "__main__":
    main()



import asyncio
from dataclasses import dataclass, field
from datetime import date, datetime, time, timezone
from enum import Enum, auto
from pathlib import Path
from unittest.mock import MagicMock
from uuid import UUID, uuid4

import pytest

from trame_server import Server
from trame_server.utils.typed_state import (
    DefaultEncoderDecoder,
    IStateEncoderDecoder,
    TypedState,
)


@pytest.fixture
def server():
    return Server()


@pytest.fixture
def state(server):
    server.state.ready()
    return server.state


@pytest.fixture
def child_state(server):
    child_server = server.create_child_server(prefix="child_")
    child_server.state.ready()
    return child_server.state


@dataclass
class MyData:
    a: int = 1
    b: int = 2


@dataclass
class MyBiggerData:
    my_other_data: MyData = field(default_factory=MyData)
    c: float = 42.0


def test_can_be_constructed_from_simple_dataclass(state):
    typed_state = TypedState(state, MyData)

    assert typed_state.data.a == 1
    typed_state.data.a = 42
    assert state[typed_state.name.a] == 42


def test_can_be_constructed_from_nested_dataclass(state):
    typed_state = TypedState(state, MyBiggerData)

    assert typed_state.data.my_other_data.a == 1
    typed_state.data.my_other_data.a = 42
    assert state[typed_state.name.my_other_data.a] == 42


def test_can_handle_namespace(state):
    typed_state_ns1 = TypedState(state, MyData, namespace="ns1")
    typed_state_ns2 = TypedState(state, MyData, namespace="ns2")

    typed_state_ns1.data.a = 40
    typed_state_ns2.data.a = 3

    assert typed_state_ns2.data.a != typed_state_ns1.data.a


def test_can_be_converted_to_dataclass(state):
    typed_state = TypedState(state, MyBiggerData)
    typed_state.data.my_other_data.a = 42
    typed_state.data.my_other_data.b = 43
    typed_state.data.c = 44

    assert typed_state.get_dataclass() == MyBiggerData(
        my_other_data=MyData(a=42, b=43), c=44
    )
    assert TypedState.as_dataclass(typed_state.data.my_other_data) == MyData(a=42, b=43)


def test_can_be_set_from_data_class(state):
    typed_state = TypedState(state, MyBiggerData)

    typed_state.set_dataclass(MyBiggerData(my_other_data=MyData(a=42, b=43), c=44))
    assert typed_state.data.my_other_data.a == 42
    assert typed_state.data.my_other_data.b == 43
    assert typed_state.data.c == 44

    TypedState.from_dataclass(typed_state.data.my_other_data, MyData(a=45, b=46))
    assert typed_state.data.my_other_data.a == 45
    assert typed_state.data.my_other_data.b == 46


def test_can_be_used_to_connect_to_state_changes(state):
    typed_state = TypedState(state, MyBiggerData, namespace="ns1")

    mock = MagicMock()

    @state.change(typed_state.name.my_other_data.a)
    def on_a_change(**_):
        mock(typed_state.data.my_other_data.a)

    typed_state.data.my_other_data.a = 53
    state.flush()

    mock.assert_called_once_with(53)


def test_can_handle_child_server(state, child_state):
    typed_state_main = TypedState(state, MyData)
    typed_state_child = TypedState(child_state, MyData)

    typed_state_main.data.a = 40
    typed_state_child.data.a = 3

    assert typed_state_main.data.a != typed_state_child.data.a
    assert typed_state_child.data.a == state[f"child_{typed_state_child.name.a}"]


class MyEnum(Enum):
    A = auto()
    B = auto()
    C = auto()


@dataclass
class DataWithTypes:
    my_enum: MyEnum
    my_uuid: UUID
    my_enum_tuple: tuple[MyEnum]
    my_enum_list: list[MyEnum]
    my_enum_dict: dict[MyEnum, MyEnum]
    my_datetime: datetime
    my_date: date
    my_time: time
    my_path: Path


def test_has_default_encoders_and_decoders_for_basic_types(state):
    typed_state = TypedState(state, DataWithTypes)
    typed_state.data.my_enum = MyEnum.B
    uuid = uuid4()
    typed_state.data.my_uuid = uuid
    typed_state.data.my_enum_tuple = (MyEnum.A, MyEnum.C)
    typed_state.data.my_enum_list = [MyEnum.A, MyEnum.C]
    typed_state.data.my_enum_dict = {MyEnum.A: MyEnum.B}

    dt = datetime.now(tz=timezone.utc)
    typed_state.data.my_datetime = dt
    typed_state.data.my_date = dt.date()
    typed_state.data.my_time = dt.time()
    typed_state.data.my_path = Path(__file__)

    assert typed_state.data.my_enum == MyEnum.B
    assert typed_state.data.my_uuid == uuid
    assert typed_state.data.my_enum_tuple == (MyEnum.A, MyEnum.C)
    assert typed_state.data.my_enum_list == [MyEnum.A, MyEnum.C]
    assert typed_state.data.my_enum_dict == {MyEnum.A: MyEnum.B}
    assert typed_state.data.my_datetime == dt
    assert typed_state.data.my_date == dt.date()
    assert typed_state.data.my_time == dt.time()
    assert typed_state.data.my_path == Path(__file__)

    assert state[typed_state.name.my_enum] == MyEnum.B.value
    assert state[typed_state.name.my_uuid] == str(uuid)
    assert state[typed_state.name.my_enum_tuple] == (MyEnum.A.value, MyEnum.C.value)
    assert state[typed_state.name.my_enum_list] == [MyEnum.A.value, MyEnum.C.value]
    assert state[typed_state.name.my_enum_dict] == {MyEnum.A.value: MyEnum.B.value}
    assert state[typed_state.name.my_datetime] == dt.isoformat()
    assert state[typed_state.name.my_date] == dt.date().isoformat()
    assert state[typed_state.name.my_time] == dt.time().isoformat()
    assert state[typed_state.name.my_path] == Path(__file__).as_posix()


def test_can_customize_encoders_and_decoders(state):
    class MyEncoder(IStateEncoderDecoder):
        def decode(self, _obj, _obj_type: type):
            return 42

        def encode(self, obj):
            return str(obj)

    typed_state = TypedState(state, DataWithTypes, encoders=[MyEncoder()])
    typed_state.data.my_enum = MyEnum.B
    assert state[typed_state.name.my_enum] == str(MyEnum.B)
    assert typed_state.data.my_enum == 42


@dataclass
class MyDataWithFactory:
    a: list[MyEnum] = field(default_factory=lambda: [MyEnum.A, MyEnum.B])


def test_is_compatible_with_default_factory(state):
    typed_state = TypedState(state, MyDataWithFactory)
    assert state[typed_state.name.a] == [MyEnum.A.value, MyEnum.B.value]


def test_different_data_classes_with_same_name_are_not_mangled_by_default(state):
    @dataclass
    class A:
        a: int = 42

    @dataclass
    class B:
        a: float = 32.3

    a = TypedState(state, A)
    b = TypedState(state, B)
    a.data.a = 3
    assert b.data.a == 32.3


def test_state_field_is_consistent_for_nested_proxies(state):
    typed_state = TypedState(state, MyBiggerData)
    data_fields = TypedState.get_field_proxy_dict(typed_state.data)

    # root data fields contains : data, data.c, data.my_other_data, data.my_other_data.a, my_other_data.b
    assert len(data_fields) == 5

    # inner field contains: my_other_data, my_other_data.a, my_other_data.b
    assert len(TypedState.get_field_proxy_dict(typed_state.data.my_other_data)) == 3


def test_can_bind_state_names_to_strongly_typed_state_callback(state):
    typed_state = TypedState(state, DataWithTypes)

    mock = MagicMock()
    typed_state.bind_changes(
        {(typed_state.name.my_date, typed_state.name.my_time): mock}
    )

    dt = datetime.now(tz=timezone.utc)

    typed_state.data.my_date = dt.date()
    typed_state.data.my_time = dt.time()

    state.flush()
    mock.assert_called_once_with(dt.date(), dt.time())


def test_can_bind_state_names_to_inner_dataclass_types(state):
    typed_state = TypedState(state, MyBiggerData)

    mock = MagicMock()
    typed_state.bind_changes(
        {(typed_state.name.my_other_data, typed_state.name.c): mock}
    )
    typed_state.data.my_other_data.a = 54

    state.flush()
    mock.assert_called_once_with(typed_state.data.my_other_data, typed_state.data.c)


def test_can_bind_to_full_typed_state(state):
    typed_state = TypedState(state, MyBiggerData)

    mock = MagicMock()
    typed_state.bind_changes({typed_state.name: mock})
    typed_state.data.my_other_data.a = 54
    state.flush()

    mock.assert_called_once_with(typed_state.data)


def test_data_proxies_key_binding_is_unpacked_for_inner_states(state):
    typed_state = TypedState(state, MyBiggerData)
    state_keys = TypedState.get_reactive_state_id_keys(
        [typed_state.name.my_other_data, typed_state.name.c]
    )
    assert state_keys == [
        typed_state.name.my_other_data.a,
        typed_state.name.my_other_data.b,
        TypedState.get_state_id(typed_state.name.my_other_data),
        typed_state.name.c,
    ]


def test_data_proxies_value_key_is_inner_state_id_for_inner_states(state):
    typed_state = TypedState(state, MyBiggerData)

    value_keys = TypedState.get_value_state_keys(
        [typed_state.data, typed_state.data.my_other_data]
    )
    assert value_keys == [
        TypedState.get_state_id(typed_state.data),
        TypedState.get_state_id(typed_state.data.my_other_data),
    ]

    field_dict = TypedState.get_field_proxy_dict(typed_state.data)
    assert all(k in field_dict for k in value_keys)


def test_can_bind_to_arbitrary_callback_args(state):
    typed_state = TypedState(state, MyBiggerData)

    mock_b = MagicMock()

    def my_callback(arbitrary_a, other_b):
        mock_b(arbitrary_a, other_b)

    typed_state.bind_changes(
        {(typed_state.name.my_other_data.b, typed_state.name.c): my_callback}
    )

    typed_state.data.my_other_data.b = 2
    typed_state.data.c = 3.4
    state.flush()

    mock_b.assert_called_once_with(2, 3.4)


class CustomEnumEncode(DefaultEncoderDecoder):
    def encode(self, obj):
        if isinstance(obj, MyEnum):
            return obj.name + "_CUSTOM"
        return super().encode(obj)

    def decode(self, obj, obj_type: type):
        if issubclass(obj_type, MyEnum) and isinstance(obj, str):
            return MyEnum[obj.replace("_CUSTOM", "")]
        return super().decode(obj, obj_type)


def test_can_encode_values_consistently_with_proxy_encoding(state):
    typed_state = TypedState(
        state,
        MyBiggerData,
        encoders=[CustomEnumEncode()],
    )

    assert typed_state.encode([{"text": e.name, "value": e} for e in MyEnum]) == [
        {"text": e.name, "value": e.name + "_CUSTOM"} for e in MyEnum
    ]


def test_can_check_if_is_specific_proxy_type(state):
    typed_state = TypedState(state, MyBiggerData)
    assert TypedState.is_name_proxy_class(typed_state.name.my_other_data)
    assert TypedState.is_data_proxy_class(typed_state.data.my_other_data)


@dataclass
class DualState:
    d1: MyData = field(default_factory=MyData)
    d2: MyData = field(default_factory=MyData)


def test_supports_creating_sub_states(state):
    typed_state = TypedState(state, DualState)

    sub_state_1 = typed_state.get_sub_state(typed_state.name.d1)
    assert isinstance(sub_state_1, TypedState)
    assert sub_state_1.name.a == typed_state.name.d1.a

    typed_state.data.d1.a = 808
    assert sub_state_1.data.a == 808

    sub_state_2 = typed_state.get_sub_state(typed_state.name.d2)
    assert sub_state_2.name.a == typed_state.name.d2.a

    assert sub_state_1._encoder == sub_state_2._encoder == typed_state._encoder


@pytest.mark.asyncio
async def test_is_compatible_with_async_bind_changes(state):
    typed_state = TypedState(state, MyData)

    mock = MagicMock()

    async def async_callback(value):
        await asyncio.sleep(0.1)
        mock(value)

    typed_state.bind_changes({typed_state.name.a: async_callback})

    with state:
        typed_state.data.a = 43

    await asyncio.sleep(0.05)
    mock.assert_not_called()
    await asyncio.sleep(0.1)
    mock.assert_called_once_with(43)


@dataclass
class DataWithUnionTypes:
    my_optional_enum: MyEnum | None
    my_str_or_enum: str | MyEnum
    my_enum_or_str: MyEnum | str
    my_optional_enum_list: list[MyEnum | str | None] | None


def test_handles_union_types_by_order_of_definition(state):
    encode = CustomEnumEncode()
    typed_state = TypedState(state, DataWithUnionTypes, encoders=[encode])

    b_enum = MyEnum.B
    b_str = encode.encode(b_enum)

    typed_state.data.my_optional_enum = None
    typed_state.data.my_str_or_enum = b_str
    typed_state.data.my_enum_or_str = b_str

    assert typed_state.data.my_optional_enum is None
    assert typed_state.data.my_str_or_enum == b_str
    assert typed_state.data.my_enum_or_str == b_enum

    assert state[typed_state.name.my_optional_enum] is None
    assert state[typed_state.name.my_str_or_enum] == b_str
    assert state[typed_state.name.my_enum_or_str] == b_str

    typed_state.data.my_optional_enum = b_enum
    typed_state.data.my_str_or_enum = b_enum
    typed_state.data.my_enum_or_str = b_enum
    assert typed_state.data.my_optional_enum == b_enum
    assert typed_state.data.my_str_or_enum == b_str
    assert typed_state.data.my_enum_or_str == b_enum

    assert state[typed_state.name.my_optional_enum] == b_str
    assert state[typed_state.name.my_str_or_enum] == b_str
    assert state[typed_state.name.my_enum_or_str] == b_str


def test_handles_list_of_union(state):
    encode = CustomEnumEncode()
    typed_state = TypedState(state, DataWithUnionTypes, encoders=[encode])

    b_enum = MyEnum.B
    b_str = encode.encode(b_enum)

    typed_state.data.my_optional_enum_list = None
    assert typed_state.data.my_optional_enum_list is None

    typed_state.data.my_optional_enum_list = [b_str, b_enum, None]
    assert typed_state.data.my_optional_enum_list == [b_enum, b_enum, None]
    assert state[typed_state.name.my_optional_enum_list] == [b_str, b_str, None]


def test_failure_to_encode_raises_type_error(state):
    class RaiseEncode(DefaultEncoderDecoder):
        def encode(self, _obj):
            raise AssertionError()

        def decode(self, _obj, _obj_type: type):
            raise AssertionError()

    typed_state = TypedState(state, DataWithTypes, encoders=[RaiseEncode()])
    state.setdefault(typed_state.name.my_enum, MyEnum.A.value)

    with pytest.raises(TypeError):
        typed_state.data.my_enum = MyEnum.A

    with pytest.raises(TypeError):
        print(typed_state.data.my_enum)


@dataclass
class SimpleTypes:
    my_int: int
    my_enum: MyEnum
    my_path: Path


@dataclass
class TypedComposite:
    simple_types: SimpleTypes = field(default_factory=SimpleTypes)


@dataclass
class DataclassCollections:
    nested_list: list[TypedComposite] = field(default_factory=list)
    nested_dict: dict[str, TypedComposite] = field(default_factory=dict)


def test_encode_decode_supports_collections_of_nested_dataclass(state):
    typed_state = TypedState(state, DataclassCollections)
    typed_state.data.nested_list = [
        TypedComposite(
            SimpleTypes(my_int=1, my_enum=MyEnum.A, my_path=Path("/path/to/1"))
        ),
        TypedComposite(
            SimpleTypes(my_int=2, my_enum=MyEnum.B, my_path=Path("/path/to/2"))
        ),
    ]

    typed_state.data.nested_dict = {
        "3": TypedComposite(
            SimpleTypes(my_int=3, my_enum=MyEnum.C, my_path=Path("/path/to/3"))
        ),
        "4": TypedComposite(
            SimpleTypes(my_int=4, my_enum=MyEnum.A, my_path=Path("/path/to/4"))
        ),
    }

    assert typed_state.data.nested_list[0] == TypedComposite(
        SimpleTypes(my_int=1, my_enum=MyEnum.A, my_path=Path("/path/to/1"))
    )
    assert typed_state.data.nested_list[1] == TypedComposite(
        SimpleTypes(my_int=2, my_enum=MyEnum.B, my_path=Path("/path/to/2"))
    )
    assert typed_state.data.nested_dict["3"] == TypedComposite(
        SimpleTypes(my_int=3, my_enum=MyEnum.C, my_path=Path("/path/to/3"))
    )
    assert typed_state.data.nested_dict["4"] == TypedComposite(
        SimpleTypes(my_int=4, my_enum=MyEnum.A, my_path=Path("/path/to/4"))
    )

    assert state[typed_state.name.nested_list] == [
        {
            "simple_types": {
                "my_int": 1,
                "my_enum": typed_state.encode(MyEnum.A),
                "my_path": typed_state.encode("/path/to/1"),
            }
        },
        {
            "simple_types": {
                "my_int": 2,
                "my_enum": typed_state.encode(MyEnum.B),
                "my_path": typed_state.encode("/path/to/2"),
            }
        },
    ]

    assert state[typed_state.name.nested_dict] == {
        "3": {
            "simple_types": {
                "my_int": 3,
                "my_enum": typed_state.encode(MyEnum.C),
                "my_path": typed_state.encode("/path/to/3"),
            }
        },
        "4": {
            "simple_types": {
                "my_int": 4,
                "my_enum": typed_state.encode(MyEnum.A),
                "my_path": typed_state.encode("/path/to/4"),
            }
        },
    }


def test_can_supress_change_listeners(state):
    mock = MagicMock()

    typed_state = TypedState(state, MyBiggerData)
    typed_state.bind_changes({typed_state.name.my_other_data: mock})
    state.flush()

    # Typed state definition sets defaults at creation.
    # Reset mock to only check later mutation.
    mock.reset_mock()

    with typed_state.suppress_change_listeners():
        typed_state.data.my_other_data.a = 53

    state.flush()
    mock.assert_not_called()



from trame.app import asynchronous, get_server
from trame.ui.vuetify import SinglePageLayout
from trame.widgets import vuetify

from trame_server.client import Client

server = get_server(client_type="vue2")
state, ctrl = server.state, server.controller
server.cli.add_argument("--url")

client = Client()

state.trame_title = "Client"
state.counter = 0


@state.change("counter")
def on_change(counter, **_):
    if counter == client.state.counter:
        return

    msg = f"client::counter = {counter}"
    if state.log is None:
        state.log = ""
    state.log += msg + "\n"
    print(msg)
    # Push local state to remote server
    with client.state:
        client.state.counter = counter


async def trigger(*args, **kwargs):
    msg = f"client::trigger = {args} {kwargs}"
    if state.log is None:
        state.log = ""
    state.log += msg + "\n"
    print(msg)
    resp = await client.call_trigger("my_method", args, kwargs)
    print("Server replied with:", resp)


@client.state.change("counter")
def on_remote_change(counter, **_):
    if counter == state.counter:
        return

    with state:
        msg = f"remote::counter = {counter}"
        if state.log is None:
            state.log = ""
        state.log += msg + "\n"
        print(msg)

        # Sync local state
        state.counter = counter


@client.state.change("test_file")
def on_file(test_file, **_):
    with state:
        msg = f"remote::test_file = {test_file.get('name')}"
        print(list(test_file.keys()))
        state.log += msg + "\n"
        print(msg)


@client.state.change("attch_data")
def on_attachment(attch_data, **_):
    with state:
        msg = f"remote::attch_data {attch_data.get('time')}"
        print(msg)
        state.log += msg + "\n"


@ctrl.add("on_server_ready")
def connect_to_server(**_):
    url = server.cli.parse_args().url
    print(f"Client connect to {url}")
    asynchronous.create_task(client.connect(url, secret="wslink-secret"))


with SinglePageLayout(server) as layout:
    layout.title.set_text(state.trame_title)

    with layout.toolbar as toolbar:
        vuetify.VSpacer()
        toolbar.add_child("{{ counter }}")
        vuetify.VSpacer()
        vuetify.VBtn("-", click="counter--")
        vuetify.VBtn("+", click="counter++")
        vuetify.VBtn(
            "fn", click=(trigger, "[counter, 2]", "{a: 2 * counter, b: 3 * counter}")
        )

    with layout.content:
        with vuetify.VContainer(fluid=True):
            vuetify.VTextarea(
                v_model=("log", ""),
                outlined=True,
                clearable=True,
                auto_grow=True,
            )

if __name__ == "__main__":
    server.start()



# Configuration file for the Sphinx documentation builder.

# -- Project information -----------------------------------------------------

project = "Trame Server"
copyright = "2022, Kitware"
author = "Kitware"

# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"



from pathlib import Path

from trame.app import get_server
from trame.ui.html import DivLayout
from trame.widgets import html

server = get_server()
state = server.state

server.enable_module(
    {
        "serve": {
            "__test": str(Path(__file__).parent.resolve()),
        },
        "scripts": ["__test/stateUpdateListener.js"],
    }
)

state.msg = "Something"
state.change_count = 0


def update_state():
    server.clear_state_client_cache("msg")
    state.dirty("msg")


with DivLayout(server):
    html.Button("A", click="msg = 'something A'")
    html.Button("B", click="msg = 'something B'")
    html.Button("Force", click=update_state)
    html.Div("MSG={{ msg }} - Change count={{ change_count }}")

server.start()



import asyncio
from unittest.mock import MagicMock

import pytest
import pytest_asyncio
from trame.app import asynchronous, get_client, get_server


@pytest_asyncio.fixture
async def server():
    server = get_server("test_client_connection")
    server.start(exec_mode="task", port=0)
    assert await server.ready
    assert server.running
    try:
        yield server
    finally:
        await asyncio.sleep(0.5)
        await server.stop()


@pytest_asyncio.fixture
async def client(server):
    url = f"ws://localhost:{server.port}/ws"
    client = get_client(url)
    asynchronous.create_task(client.connect(secret="wslink-secret"))

    for _ in range(10):
        if client.connected == 2:
            break
        await asyncio.sleep(0.1)

    # should be a noop
    await client.connect()
    assert client.connected == 2

    try:
        yield client
    finally:
        await asyncio.sleep(0.1)
        await client.disconnect()


@pytest.mark.asyncio
async def test_client_connection(server, client):
    @client.change("a")
    def on_change(a, **_):
        assert a == 2

    @server.trigger("add")
    def server_method(*args):
        result = 0
        for v in args:
            result += v
        return result

    with server.state as state:
        state.a = 2

    await server.network_completion
    await asyncio.sleep(0.1)  # wait for client network

    assert server.state.a == client.state.a

    with client.state as state:
        state.b = {"a": 1, "b": 2, "_filter": ["b"]}

    await asyncio.sleep(0.1)  # wait for client network
    assert server.state.b == {"a": 1, "_filter": ["b"]}

    assert await client.call_trigger("add", [1, 2, 3]) == 6
    assert await client.call_trigger("add") == 0


@pytest.mark.asyncio
async def test_given_suppress_listeners_sends_updates_to_client_only(server, client):
    state_key = "my_state_key"

    on_call = MagicMock()
    server.state.change(state_key)(on_call)
    with server.state.suppress_change_listeners(state_key):
        server.state[state_key] = 42
        server.state.flush()

    await server.network_completion
    await asyncio.sleep(0.1)

    assert client.state[state_key] == 42
    on_call.assert_not_called()



from trame.app import get_server
from trame.ui.vuetify import SinglePageLayout
from trame.widgets import vuetify2 as vuetify

server = get_server(client_type="vue2")
state, ctrl = server.state, server.controller

state.field1 = 1
state.field2 = 2
state.field3 = 3

VAR_NAMES = ["field1", "field2", "field3"]


@state.change(*VAR_NAMES)
def change_detected(**_):
    print(f"Triggered because {list(state.modified_keys)} changed")
    for n in state.modified_keys:
        print(f"{n} = {state[n]}")


with SinglePageLayout(server) as layout:
    with layout.content:
        with vuetify.VContainer(
            fluid=True,
            classes="d-flex justify-center align-center",
            style="height: 100%;",
        ):
            with vuetify.VCard(style="max-width: 400px;"):
                vuetify.VCardTitle("VCard")
                with vuetify.VCardText():
                    vuetify.VTextField(
                        v_model=("field1",),
                        label="Field 1",
                        outlined=True,
                    )
                    vuetify.VTextField(
                        v_model=("field2",),
                        label="Field 2",
                        outlined=True,
                    )
                    vuetify.VTextField(
                        v_model=("field3",),
                        label="Field 3",
                        outlined=True,
                    )

server.start()



from trame.app import get_server
from trame.ui.html import DivLayout
from trame.widgets import html


def test_namespace_template():
    server = get_server("test_namespace_template")
    child_server = server.create_child_server(prefix="child_")
    child_server.state.a = 10

    layout = DivLayout(child_server)
    with layout:
        html.Div("{{ a }}")

    assert layout.html == "<div >\n<div >\n{{ child_a }}\n</div>\n</div>"
    assert child_server.translator("a") == "child_a"



import random

from trame.app import get_server
from trame.decorators import TrameApp, change, trigger
from trame.ui.vuetify3 import SinglePageLayout
from trame.widgets import vuetify3 as v3


def build_dialog():
    with v3.VDialog(v_model=("dialog_show", False), width=300) as d:
        v3.VCard(title="Test", text="This is a dialog.")

        # force state default at exec so translation can work
        d.html  # noqa: B018


@TrameApp()
class MainApp:
    def __init__(self, server=None):
        self.server = get_server(server)
        self._build_ui()

        SecondApp(self.server)

    def _build_ui(self):
        with SinglePageLayout(self.server, full_height=True) as layout:
            with layout.content:
                build_dialog()
                v3.VBtn(
                    "Open Second App",
                    click="window.open('/?ui=second', target='_blank')",
                )
                v3.VBtn("Test Trigger", click="trame.trigger('test')")
                v3.VBtn("Unique Main Trigger", click="trame.trigger('main')")
                v3.VBtn("Show dialog", click="dialog_show = true")

    @trigger("test")
    def test_trigger(self):
        print("test main")

    @trigger("main")
    def main_trigger(self):
        print("unique main")


@TrameApp()
class SecondApp:
    def __init__(self, server=None, template_name="second"):
        self.prefix = ""
        if server:
            self.prefix = "second_"
            self.server = server.create_child_server(prefix=self.prefix)
        else:
            self.server = get_server(server)

        self.state = self.server.state
        self.ctrl = self.server.controller
        self.translator = self.server.translator

        self.state.random_value = random.random()

        # timing thing to get name resolved
        self.server.trigger("second")(self.second_trigger)

        self._build_ui(template_name)

    @trigger("test")
    def test_trigger(self):
        print("test second")

    @trigger("second")
    def second_trigger(self):
        print("unique second")

    @change("random_value")
    def random_value_changed(self, **_):
        print(f"Random value changed to: {self.state.random_value}")

    def random_print(self, random_value):
        print(f"{random_value=}")

    def _build_ui(self, template_name):
        # self.state.dialog_show = False
        with SinglePageLayout(
            self.server, template_name=template_name, full_height=True
        ) as layout:
            with layout.toolbar.clear() as toolbar:
                toolbar.density = "compact"
                toolbar.title = "Data Table Example"

            with layout.content:
                build_dialog()
                v3.VBtn("Test Trigger", click=f"trame.trigger('{self.prefix}test')")
                v3.VBtn("Test Trigger ctrl", click=self.test_trigger)
                v3.VBtn(
                    "Unique Second Trigger",
                    click=f"trame.trigger('{self.prefix}second')",
                )
                v3.VBtn("Unique Second Trigger ctrl", click=self.second_trigger)
                v3.VBtn(
                    "Unique Second Trigger fix",
                    click=f"trame.trigger('{self.ctrl.trigger_name(self.second_trigger)}')",
                )
                v3.VBtn(
                    "Show dialog",
                    click=self.translator.translate_js_expression(
                        self.state, "dialog_show = true"
                    ),
                )
                v3.VBtn(
                    "Change random_value",
                    click=self.translator.translate_js_expression(
                        self.state, "random_value = Math.random()"
                    ),
                )
                v3.VBtn(
                    "Change random_value (auto translate)",
                    click="random_value = Math.random()",
                )
                v3.VBtn(
                    "Change random_value (arg auto translate)",
                    click=(self.random_print, "[random_value]"),
                )


def main():
    app = MainApp()
    app.server.start()


if __name__ == "__main__":
    main()



import asyncio
import logging
import os
import traceback

import aiohttp
import msgpack
from wslink.chunking import UnChunker, generate_chunks

from trame_server.utils import asynchronous

from .state import State

MAX_MSG_SIZE = int(os.environ.get("WSLINK_MAX_MSG_SIZE") or 4194304)

logger = logging.getLogger(__name__)


class WsLinkSession:
    CLIENT_ERROR = -32099
    AUTH_ID = "system:c0:0"

    def __init__(self, ws):
        self.loop = asyncio.get_running_loop()
        self.attachment_atomic = asyncio.Lock()
        self.ws = ws
        self.msg_count = 0
        self.bin_id = 1
        self.subscriptions = {}
        self.client_id = None
        self.unchunker = UnChunker()
        self.in_flight_rpc = {}

    async def on_msg_complete(self, payload):
        # Notification-only message from the server - should be binary attachment header
        if "id" not in payload:
            return

        msg_id = payload.get("id")
        msg_type, msg_topic, _ = msg_id.split(":")
        future = self.in_flight_rpc.get(msg_id)

        # Error
        if "error" in payload:
            if future:
                future.set_exception(
                    payload.get("error", "Server error")
                )  # May need to wrap in Exception?
            else:
                print("Server error:", payload.get("error"))

            self.in_flight_rpc.pop(msg_id)
            return

        # Normal processing
        msg_result = payload.get("result")

        # RPC
        if msg_type == "rpc":
            if future:
                future.set_result(msg_result)

        # Publish
        if msg_type == "publish" and msg_topic in self.subscriptions:
            event = msg_result
            for fn in self.subscriptions[msg_topic]:
                try:
                    fn(event)
                except Exception:
                    print("Subscription callback error")
                    traceback.print_exc()

        # System
        if msg_type == "system":
            if msg_id == WsLinkSession.AUTH_ID:
                self.client_id = msg_result.get("clientID")
                self.unchunker.set_max_message_size(msg_result.get("maxMsgSize"))
                future.set_result(self.client_id)
            else:
                future.set_result(msg_result)

        # Clean pending future
        if future:
            self.in_flight_rpc.pop(msg_id)

    async def listen(self):
        async for msg in self.ws:
            if msg.type == aiohttp.WSMsgType.CLOSE:
                print("CLOSE")
            elif msg.type == aiohttp.WSMsgType.CLOSING:
                print("CLOSING")
            elif msg.type == aiohttp.WSMsgType.CLOSED:
                print("CLOSED")
            elif msg.type == aiohttp.WSMsgType.ERROR:
                print("ERROR")
            elif msg.type == aiohttp.WSMsgType.TEXT:
                logger.critical("wslink is not expecting text message:\n> %s", msg.data)
            if msg.type == aiohttp.WSMsgType.BINARY:
                full_message = self.unchunker.process_chunk(msg.data)
                if full_message is not None:
                    await self.on_msg_complete(full_message)

    async def auth(self, **kwargs):
        key = WsLinkSession.AUTH_ID
        resp = self.loop.create_future()
        wrapper = {
            "wslink": "1.0",
            "id": key,
            "method": "wslink.hello",
            "args": [kwargs],
            "kwargs": {},
        }
        self.in_flight_rpc[key] = resp

        try:
            packed_wrapper = msgpack.packb(wrapper)
        except Exception:
            del wrapper["error"]["data"]
            packed_wrapper = msgpack.packb(wrapper)

        async with self.attachment_atomic:
            for chunk in generate_chunks(packed_wrapper, MAX_MSG_SIZE):
                if self.ws is not None:
                    await self.ws.send_bytes(chunk)

        return resp

    async def call(self, method, args=None, kwargs=None):
        self.msg_count += 1
        key = f"rpc:{self.client_id}:{self.msg_count}"
        resp = self.loop.create_future()
        self.in_flight_rpc[key] = resp
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        wrapper = {
            "wslink": "1.0",
            "id": key,
            "method": method,
            "args": args,
            "kwargs": kwargs,
        }

        try:
            packed_wrapper = msgpack.packb(wrapper)
        except Exception:
            del wrapper["error"]["data"]
            packed_wrapper = msgpack.packb(wrapper)

        async with self.attachment_atomic:
            for chunk in generate_chunks(packed_wrapper, MAX_MSG_SIZE):
                if self.ws is not None:
                    await self.ws.send_bytes(chunk)

        return resp

    def register_subscription(self, topic, callback):
        if topic not in self.subscriptions:
            self.subscriptions[topic] = [callback]
        else:
            self.subscriptions[topic].append(callback)

    def unregister_subscription(self, topic, callback):
        callbacks = self.subscriptions.get(topic, [])
        if callback in callbacks:
            callbacks.remove(callback)

        if len(callbacks) == 0 and topic in self.subscriptions:
            self.subscriptions.pop(topic)

    def clear_subscriptions(self):
        topics = list(self.subscriptions.keys())
        for topic in topics:
            self.subscriptions.pop(topic)

    async def close(self):
        if self.ws:
            await self.ws.close()


class Client:
    """
    Client implementation for driving a remote trame server with its shared state and
    trigger method calls in plain python.
    """

    def __init__(self, url=None, config=None, translator=None, hot_reload=False):
        # Network
        self._connected = 0
        self._session = None
        self._url = url
        self._config = {} if config is None else config

        # fake server
        self.hot_reload = hot_reload
        self._change_callbacks = {}

        # trame state
        self._state = State(
            translator, commit_fn=self._push_state, hot_reload=hot_reload
        )

    async def connect(self, url=None, **kwargs):
        if self._connected:
            return
        self._connected = 1

        config = {**self._config, **kwargs}
        if url is None:
            url = self._url

        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(url) as ws:
                self._session = WsLinkSession(ws)
                self._state.ready()
                self._session.register_subscription(
                    "trame.state.topic", self._on_state_update
                )
                self._connected = 2
                task = asynchronous.create_task(self._session.listen())
                await self._session.auth(**config)
                await task

        self._session.clear_subscriptions()
        self._session = None
        self._connected = 0

    async def disconnect(self):
        if self._session:
            await self._session.close()

    # -----------------------------------------------------
    # Fake server for state
    # -----------------------------------------------------

    @property
    def change(self):
        """
        Use as decorator `@server.change(key1, key2, ...)` so the decorated function
        will be called like so `_fn(**state)` when any of the listed key name
        is getting modified from either client or server.

        :param *_args: A list of variable name to monitor
        :type *_args: str
        """
        return self._state.change

    def _push_state(self, state):
        if self._session and self._session.client_id is not None:
            delta = []
            for key, value in state.items():
                if isinstance(value, dict) and "_filter" in value:
                    skip_keys = set(value.get("_filter"))
                    new_value = {}
                    for k, v in value.items():
                        if k not in skip_keys:
                            new_value[k] = v
                    delta.append({"key": key, "value": new_value})
                else:
                    delta.append({"key": key, "value": value})
            asynchronous.create_task(self._session.call("trame.state.update", [delta]))

    def _on_state_update(self, modified_state):
        with self.state:
            self.state.update(modified_state)

    # -----------------------------------------------------

    @property
    def connected(self):
        return self._connected

    @property
    def state(self):
        return self._state

    async def call_trigger(self, name, args=None, kwargs=None):
        if args is None:
            args = []

        if kwargs is None:
            kwargs = {}

        response = await self._session.call("trame.trigger", [name, args, kwargs])
        return await response



import logging

from trame_server.core import Controller, State, Translator

logger = logging.getLogger(__name__)


def func1():
    return 1


def func2():
    return 2


def func3():
    return 3


def test_translator():
    a_translator = Translator()
    a_translator.add_translation("foo", "a_foo")

    assert a_translator.translate_key("foo") == "a_foo"
    assert a_translator.translate_key("bar") == "bar"
    assert a_translator.reverse_translate_key("a_foo") == "foo"
    assert a_translator.reverse_translate_key("bar") == "bar"

    b_translator = Translator()
    b_translator.set_prefix("b_")

    assert b_translator.translate_key("foo") == "b_foo"
    assert b_translator.translate_key("bar") == "b_bar"
    assert b_translator.reverse_translate_key("b_foo") == "foo"
    assert b_translator.reverse_translate_key("b_bar") == "bar"

    c_translator = Translator()
    c_translator.set_prefix("c_")
    c_translator.add_translation("foo", "still_foo")

    assert c_translator.translate_key("foo") == "still_foo"
    assert c_translator.translate_key("bar") == "c_bar"
    assert c_translator.reverse_translate_key("still_foo") == "foo"
    assert c_translator.reverse_translate_key("c_foo") == "foo"
    assert c_translator.reverse_translate_key("c_bar") == "bar"


def test_state_translation():
    root_state = State()
    a_state = State(internal=root_state)
    b_state = State(internal=root_state)

    root_state.ready()
    a_state.ready()
    b_state.ready()

    # Since the translator doesn't have a prefix or any translation,
    # changing a piece of state on any of the 3 states applies to all
    root_state.value = 123
    assert root_state.value == 123
    assert a_state.value == 123
    assert b_state.value == 123

    a_state.value = 456
    assert root_state.value == 456
    assert a_state.value == 456
    assert b_state.value == 456

    b_state.value = 789
    assert root_state.value == 789
    assert a_state.value == 789
    assert b_state.value == 789

    # Add translations for a_state and b_state which will cause
    # "value" to point to a different key
    a_state.translator.add_translation("value", "a_value")
    b_state.translator.add_translation("value", "b_value")
    root_state.value = 123
    a_state.value = 456
    b_state.value = 789
    assert root_state.value == 123
    assert a_state.value == 456
    assert b_state.value == 789
    assert root_state.a_value == 456
    assert root_state.b_value == 789

    expected_state = {
        "value": 123,
        "a_value": 456,
        "b_value": 789,
    }
    assert expected_state == root_state.to_dict()


def test_controller_translation():
    root_controller = Controller()
    a_controller = Controller(internal=root_controller)
    b_controller = Controller(internal=root_controller)

    root_controller.func = func1
    assert root_controller.func() == 1
    assert a_controller.func() == 1
    assert b_controller.func() == 1

    a_controller.func = func2
    assert root_controller.func() == 2
    assert a_controller.func() == 2
    assert b_controller.func() == 2

    b_controller.func = func3
    assert root_controller.func() == 3
    assert a_controller.func() == 3
    assert b_controller.func() == 3

    a_controller._translator.add_translation("func", "a_func")
    b_controller._translator.add_translation("func", "b_func")
    root_controller.func = func1
    a_controller.func = func2
    b_controller.func = func3

    assert root_controller.func() == 1
    assert a_controller.func() == 2
    assert b_controller.func() == 3
    assert root_controller.a_func() == 2
    assert root_controller.b_func() == 3

    expected_controller = {
        "func": func1,
        "a_func": func2,
        "b_func": func3,
    }
    assert all(
        func in root_controller._func_dict.keys() for func in expected_controller.keys()
    )
    assert all(
        expected_controller[func]() == root_controller[func]()
        for func in expected_controller.keys()
    )


def test_state_prefix():
    root_state = State()
    a_state = State(internal=root_state)
    b_state = State(internal=root_state)

    root_state.ready()
    a_state.ready()
    b_state.ready()

    a_state.translator.set_prefix("a_")
    b_state.translator.set_prefix("b_")

    root_state.value = 123
    a_state.value = 456
    b_state.value = 789
    assert root_state.value == 123
    assert a_state.value == 456
    assert b_state.value == 789
    assert root_state.a_value == 456
    assert root_state.b_value == 789

    expected_state = {
        "value": 123,
        "a_value": 456,
        "b_value": 789,
    }
    assert expected_state == root_state.to_dict()


def test_controller_prefix():
    root_controller = Controller()
    a_controller = Controller(internal=root_controller)
    b_controller = Controller(internal=root_controller)

    a_controller._translator.set_prefix("a_")
    b_controller._translator.set_prefix("b_")
    root_controller.func = func1
    a_controller.func = func2
    b_controller.func = func3

    assert root_controller.func() == 1
    assert a_controller.func() == 2
    assert b_controller.func() == 3
    assert root_controller.a_func() == 2
    assert root_controller.b_func() == 3

    expected_controller = {
        "func": func1,
        "a_func": func2,
        "b_func": func3,
    }
    assert all(
        func in root_controller._func_dict.keys() for func in expected_controller.keys()
    )
    assert all(
        expected_controller[func]() == root_controller[func]()
        for func in expected_controller.keys()
    )


def test_state_prefix_and_translation():
    root_state = State()
    a_state = State(internal=root_state)
    b_state = State(internal=root_state)

    root_state.ready()
    a_state.ready()
    b_state.ready()

    # The states will be isolated by default
    a_state.translator.set_prefix("a_")
    b_state.translator.set_prefix("b_")

    # But by adding translations for a_state and b_state
    # that point to a common key they are still able to interact
    root_state.translator.add_translation("shared_value", "common_shared_value")
    a_state.translator.add_translation("shared_value", "common_shared_value")
    b_state.translator.add_translation("shared_value", "common_shared_value")

    root_state.value = 123
    a_state.value = 456
    b_state.value = 789
    assert root_state.value == 123
    assert a_state.value == 456
    assert b_state.value == 789
    assert root_state.a_value == 456
    assert root_state.b_value == 789

    root_state.shared_value = 123
    assert root_state.shared_value == 123
    assert a_state.shared_value == 123
    assert b_state.shared_value == 123

    a_state.shared_value = 456
    assert root_state.shared_value == 456
    assert a_state.shared_value == 456
    assert b_state.shared_value == 456

    b_state.shared_value = 789
    assert root_state.shared_value == 789
    assert a_state.shared_value == 789
    assert b_state.shared_value == 789

    expected_state = {
        "value": 123,
        "a_value": 456,
        "b_value": 789,
        "common_shared_value": 789,
    }
    assert expected_state == root_state.to_dict()


def test_controller_prefix_and_translation():
    root_controller = Controller()
    a_controller = Controller(internal=root_controller)
    b_controller = Controller(internal=root_controller)

    a_controller._translator.set_prefix("a_")
    b_controller._translator.set_prefix("b_")

    root_controller._translator.add_translation("shared_func", "common_shared_func")
    a_controller._translator.add_translation("shared_func", "common_shared_func")
    b_controller._translator.add_translation("shared_func", "common_shared_func")

    root_controller.func = func1
    a_controller.func = func2
    b_controller.func = func3
    assert root_controller.func() == 1
    assert a_controller.func() == 2
    assert b_controller.func() == 3
    assert root_controller.a_func() == 2
    assert root_controller.b_func() == 3

    root_controller.shared_func = func1
    assert root_controller.shared_func() == 1
    assert a_controller.shared_func() == 1
    assert b_controller.shared_func() == 1

    a_controller.shared_func = func2
    assert root_controller.shared_func() == 2
    assert a_controller.shared_func() == 2
    assert b_controller.shared_func() == 2

    b_controller.shared_func = func3
    assert root_controller.shared_func() == 3
    assert a_controller.shared_func() == 3
    assert b_controller.shared_func() == 3

    expected_controller = {
        "func": func1,
        "a_func": func2,
        "b_func": func3,
        "common_shared_func": func3,
    }

    assert all(
        func_name in root_controller._func_dict.keys()
        for func_name in expected_controller.keys()
    )
    assert all(
        func() == root_controller[func_name]()
        for func_name, func in expected_controller.items()
    )


def test_change_callback():
    # Ensure change callbacks are passed translated kwargs when using translations
    test_passed = False

    root_state = State()

    a_state = State(internal=root_state)
    a_state.translator.add_translation("foo", "a_foo")

    def on_a_foo_change(*_args, **kwargs):
        nonlocal test_passed
        assert "foo" in kwargs
        assert "a_foo" not in kwargs
        assert kwargs["foo"] == 123
        test_passed = "foo" in kwargs and "a_foo" not in kwargs

    a_state.change("foo")(on_a_foo_change)
    a_state.ready()
    a_state.foo = 123
    root_state.foo = 456
    a_state.flush()

    assert test_passed

    # Ensure change callbacks are passed translated kwargs when using prefix
    test_passed = False

    root_state = State()

    b_state = State(internal=root_state)
    b_state.translator.set_prefix("b_")

    def on_b_foo_change(*_args, **kwargs):
        nonlocal test_passed
        assert "foo" in kwargs
        assert "b_foo" not in kwargs
        assert kwargs["foo"] == 456
        test_passed = "foo" in kwargs and "b_foo" not in kwargs

    b_state.change("foo")(on_b_foo_change)

    b_state.ready()
    b_state.foo = 456
    root_state.foo = 123
    b_state.flush()

    assert test_passed



import nox


@nox.session(python=["3.8", "3.9", "3.10", "3.11", "3.12"])
def tests(session):
    session.install(".[dev]")
    session.install("-r", "./tests/requirements.txt")

    session.run("pytest")


@nox.session
def lint(session):
    session.install(".[dev]")
    session.run("pre-commit", "run", "--all-files")



import asyncio
import multiprocessing
import time
from concurrent.futures import ProcessPoolExecutor

import pytest
from trame.app import asynchronous, get_server


@pytest.mark.asyncio
async def test_thread_state_sync():
    running_states = []
    value_changes = []

    MULTI_PROCESS_MANAGER = multiprocessing.Manager()
    SPAWN = multiprocessing.get_context("spawn")
    PROCESS_EXECUTOR = ProcessPoolExecutor(1, mp_context=SPAWN)

    loop = asyncio.get_event_loop()
    queue = MULTI_PROCESS_MANAGER.Queue()

    server = get_server("test_thread_state_sync")
    server.state.running = False
    server.state.a = 0

    @server.state.change("running")
    def on_running_change(running, **_):
        running_states.append(running)

    @server.state.change("a")
    def on_a_change(a, **_):
        value_changes.append(a)

    server.start(exec_mode="task", port=0)
    assert await server.ready

    def exec_in_thread(queue):
        with asynchronous.StateQueue(queue) as state:
            assert state.queue is queue

            state.running = True

            state.update(
                {
                    "b": 10,
                    "c": 20,
                }
            )

            for i in range(10):
                time.sleep(0.1)
                state.a = i
                assert state.a == i
                assert state["a"] == i

            state.running = False

    asynchronous.decorate_task(
        loop.run_in_executor(
            PROCESS_EXECUTOR,
            exec_in_thread(queue),
        )
    )
    asynchronous.create_state_queue_monitor_task(server, queue)

    previous_size = len(value_changes)
    while len(value_changes) < 10:
        await asyncio.sleep(0.15)
        assert len(value_changes) > previous_size
        previous_size = len(value_changes)

    assert running_states == [False, True, False]

    assert server.state.b == 10
    assert server.state.c == 20

    await server.stop()


@pytest.mark.asyncio
async def test_task_decorator():
    bg_update = "idle"

    @asynchronous.task
    async def run_something():
        nonlocal bg_update
        bg_update = "ok"

    run_something()
    assert bg_update == "idle"
    await asyncio.sleep(0.1)
    assert bg_update == "ok"



from __future__ import annotations

import asyncio
import inspect
import logging
import os
import sys
from typing import Literal

from . import utils
from .controller import Controller
from .http import HttpHeader
from .protocol import CoreServer
from .state import State
from .ui import VirtualNodeManager
from .utils import share
from .utils.argument_parser import ArgumentParser
from .utils.namespace import Translator

logger = logging.getLogger(__name__)

ClientType = Literal["vue2", "vue3"]
BackendType = Literal["aiohttp", "generic", "tornado", "jupyter"]
ExecModeType = Literal["main", "desktop", "task", "coroutine"]

DEFAULT_CLIENT_TYPE: ClientType = "vue3"


def set_default_client_type(value: ClientType) -> None:
    global DEFAULT_CLIENT_TYPE  # noqa: PLW0603
    DEFAULT_CLIENT_TYPE = value


class Server:
    """
    Server implementation for trame.
    This is the core object that manage client/server communication but also
    holds a state and controller instance.
    With trame a server instance should be retrieved by using **trame.app.get_server()**

    Known options:
      - log_network: False (path to log file)
      - ws_max_msg_size: 10000000 (bytes)
      - ws_heart_beat: 30
      - desktop_debug: False

    :param name: A name identifier for a given server
    :type name: str, optional (default: trame)

    :param **options: Gather any keyword arguments into options
    :type options: Dict
    """

    def __init__(
        self,
        name="trame",
        vn_constructor=None,
        translator=None,
        parent_server=None,
        **options,
    ) -> None:
        # Core internal variables
        self._parent_server = parent_server
        self._translator = translator or Translator()
        self._name = share(parent_server, "_name", name)
        self._options = share(parent_server, "_options", options)
        self._client_type = share(parent_server, "_client_type", None)
        self._http_header = share(parent_server, "_http_header", HttpHeader())

        # use parent_server instead of local version
        self._server = None
        self._running_stage = 0  # 0: off / 1: pending / 2: running
        self._running_port = 0
        self._running_future = None
        self._www = None
        self.serve = {}  # HTTP static endpoints
        self._loaded_modules = set()
        self._loaded_module_dicts = []
        self._cli_parser = None
        self._root_protocol = None
        self._protocols_to_configure = []

        # ENV variable mapping settings
        self.hot_reload = "--hot-reload" in sys.argv or bool(
            os.getenv("TRAME_HOT_RELOAD", None)
        )
        if parent_server is None:
            self._options["log_network"] = self._options.get(
                "log_network", os.environ.get("TRAME_LOG_NETWORK")
            )
            self._options["ws_max_msg_size"] = self._options.get(
                "ws_max_msg_size", os.environ.get("TRAME_WS_MAX_MSG_SIZE") or 10000000
            )
            self._options["ws_heart_beat"] = self._options.get(
                "ws_heart_beat", os.environ.get("TRAME_WS_HEART_BEAT") or 30
            )
            self._options["desktop_debug"] = self._options.get(
                "desktop_debug", os.environ.get("TRAME_DESKTOP_DEBUG")
            )
            # reset default wslink startup message
            os.environ["WSLINK_READY_MSG"] = ""

        # Shared state + reserve internal keys
        if parent_server is None:
            self._state = State(
                self.translator, commit_fn=self._push_state, hot_reload=self.hot_reload
            )
            for key in ["scripts", "module_scripts", "styles", "vue_use", "mousetrap"]:
                self._state[f"trame__{key}"] = []
            self._state.trame__client_only = ["trame__busy"]
            self._state.trame__busy = 1
            self._state.trame__favicon = None
            self._state.trame__title = "Trame"
        else:
            self._state = State(
                self.translator,
                internal=parent_server._state,
                commit_fn=self._push_state,
                hot_reload=self.hot_reload,
            )

        # Controller
        if parent_server is None:
            self._controller = Controller(self.translator, hot_reload=self.hot_reload)
        else:
            self._controller = Controller(
                self.translator,
                internal=parent_server._controller,
                hot_reload=self.hot_reload,
            )

        # Server only context
        if parent_server is None:
            self._context = State(self.translator, hot_reload=self.hot_reload)
        else:
            self._context = State(
                self.translator,
                internal=parent_server._context,
                hot_reload=self.hot_reload,
            )

        # UI (FIXME): use for translator
        self._ui = share(parent_server, "_ui", VirtualNodeManager(self, vn_constructor))

    def create_child_server(self, translator=None, prefix=None) -> Server:
        translator = translator or Translator(prefix=prefix)
        return Server(translator=translator, parent_server=self)

    # -------------------------------------------------------------------------
    # State management helpers
    # -------------------------------------------------------------------------

    def _push_state(self, state):
        if self.protocol:
            self.protocol.push_state_change(state)

    # -------------------------------------------------------------------------
    # Initialization helper
    # -------------------------------------------------------------------------

    @property
    def http_headers(self):
        """Return http header helper so they can be applied before the server start."""
        return self._http_header

    def enable_module(self, module, **kwargs):
        """
        Expend server using a module definition which can be used to serve custom
        client code or assets, load/initialize resources (js, css, vue),
        register custom protocols and even execute custom code.

        Any previously seem module will be automatically skipped.

        The attributes that are getting processed in a module are the following:
          - setup(server, **kwargs): Function called first
          - scripts = []           : List all JavaScript URL that should be loaded
          - module_scripts = []    : List all JavaScript URL as type=module to load
          - styles  = []           : List all CSS URL that should be loaded
          - vue_use = ['libName', ('libName2', { **options })]: List Vue plugin to load
          - state = {}             : Set of variable to add to state
          - serve = { data: '/path/on/fs' }: Set of endpoints to serve static content
          - www = '/path/on/fs'    : Path served as main web content

        :param module: A module to enable or a dict()
        :param kwargs: Any optional parameters needed for your module setup() function.
        """
        if self.root_server != self:
            return self.root_server.enable_module(module, **kwargs)

        # Make sure definitions is a dict while skipping already loaded module
        definitions = module
        if isinstance(definitions, dict):
            if definitions in self._loaded_module_dicts:
                return False

            self._loaded_module_dicts.append(definitions)
        elif definitions in self._loaded_modules:
            return False
        else:
            self._loaded_modules.add(definitions)
            definitions = definitions.__dict__

        if "setup" in definitions:
            definitions["setup"](self, **kwargs)

        for key in ["scripts", "module_scripts", "styles", "vue_use"]:
            if key in definitions:
                self.state[f"trame__{key}"] += definitions[key]

        if "state" in definitions:
            self.state.update(definitions["state"])

        if "serve" in definitions:
            self.serve.update(definitions["serve"])

        if "www" in definitions:
            self._www = definitions["www"]

        # Reduce vue_use to merge options
        utils.reduce_vue_use(self.state)

        return True

    # -------------------------------------------------------------------------
    # Call methods
    # -------------------------------------------------------------------------

    def js_call(self, ref: str | None = None, method: str | None = None, *args):
        """
        Python call method on JS element.

        :param ref: ref name of the widget element
        :type ref: str

        :param method: name of the method that should be called
        :type method: str

        :param *args: set of parameters needed for the function
        """
        if self.protocol:
            self.protocol.push_actions(
                [
                    {
                        "type": "method",
                        "ref": ref,
                        "method": method,
                        "args": list(args),
                    },
                ],
            )

    # -------------------------------------------------------------------------
    # Annotations
    # -------------------------------------------------------------------------

    @property
    def change(self):
        """
        Use as decorator `@server.change(key1, key2, ...)` so the decorated function
        will be called like so `_fn(**state)` when any of the listed key name
        is getting modified from either client or server.

        :param *_args: A list of variable name to monitor
        :type *_args: str
        """
        return self._state.change

    # -------------------------------------------------------------------------

    @property
    def trigger(self):
        """
        Use as decorator `@server.trigger(name)` so the decorated function
        will be able to be called from the client by doing `click="trigger(name)"`.

        :param name: A name to use for that trigger
        :type name: str
        """
        return self._controller.trigger

    # -------------------------------------------------------------------------
    # From a function get its trigger name and register it if need be
    # -------------------------------------------------------------------------

    @property
    def trigger_name(self):
        """
        Given a function this method will register a trigger and returned its name.
        If manually registered, the given name at the time will be returned.

        :return: The trigger name for that function
        :rtype: str
        """
        return self._controller.trigger_name

    # -------------------------------------------------------------------------
    # App properties
    # -------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """Name of server"""
        return self._name

    @property
    def root_server(self) -> Server:
        """Root server to start"""
        if self._parent_server:
            return self._parent_server.root_server
        return self

    @property
    def translator(self):
        """Translator of the server"""
        return self._translator

    @property
    def options(self):
        """Server options provided at instantiation time"""
        return self._options

    @property
    def client_type(self) -> ClientType:
        """Specify the client type. Either 'vue2' or 'vue3' for now."""
        if self._client_type is None:
            return DEFAULT_CLIENT_TYPE  # default
        return self._client_type

    @client_type.setter
    def client_type(self, value: ClientType) -> None:
        """Should only be called once before any widget initialization"""
        if self._client_type is None:
            self._client_type = value

        if self.client_type != value:
            msg = (
                f"Trying to switch client_type from {self._client_type} to {value}."
                "The client_type can only be set once."
            )
            raise TypeError(msg)

    @property
    def cli(self):
        """argparse parser"""
        if self.root_server != self:
            return self.root_server.cli

        if self._cli_parser:
            return self._cli_parser

        self._cli_parser = ArgumentParser(
            description="Kitware trame",
            allow_abbrev=False,
        )

        # Trame specific args
        self._cli_parser.add_argument(
            "--server",
            help="Prevent your browser from opening at startup",
            action="store_true",
        )
        self._cli_parser.add_argument(
            "--banner",
            help="Print trame banner",
            action="store_true",
        )
        self._cli_parser.add_argument(
            "--app",
            help="Use OS built-in browser",
            action="store_true",
        )
        self._cli_parser.add_argument(
            "--no-http",
            help="Do not serve anything over http",
            dest="no_http",
            action="store_true",
        )
        self._cli_parser.add_argument(
            "--authKeyFile",
            help="""Path to a File that contains the Authentication key for clients
                    to connect to the WebSocket.
                    This takes precedence over '-a, --authKey' from wslink.""",
        )
        self._cli_parser.add_argument(
            "--hot-reload",
            help="""Automatically reload state/controller callback functions for every
                    function call. This allows live editing of the functions. Functions
                    located in the site-packages directories are skipped.""",
            action="store_true",
        )
        self._cli_parser.add_argument(
            "--trame-args",
            help="""If specified, trame will ignore all other arguments, and only the contents
                    of the `--trame-args` will be used. For example:
                    `--trame-args="-p 8081 --server"`. Alternatively, the environment variable
                    `TRAME_ARGS` may be set instead.""",
        )
        self._cli_parser.add_argument(
            "--follow-symlinks",
            dest="static_follow_symlinks",
            help="""flag for allowing to follow symlinks that lead outside
                    the static root directory, by default it's not allowed
                    and HTTP/404 will be returned on access.
                    Enabling follow_symlinks can be a security risk,
                    and may lead to a directory transversal attack.
                    You do NOT need this option to follow symlinks which point
                    to somewhere else within the static directory, this option
                    is only used to break out of the security sandbox.
                    Enabling this option is highly discouraged, and only
                    expected to be used for edge cases in a local development
                    setting where remote users do not have access to the server.""",
            action="store_true",
        )

        CoreServer.add_arguments(self._cli_parser)

        return self._cli_parser

    @property
    def state(self) -> State:
        """
        :return: The server shared state
        :rtype: trame_server.state.State
        """
        return self._state

    @property
    def context(self) -> State:
        """
        The server-only context (not shared with the client).

        :return: The server context state
        :rtype: trame_server.state.State
        """
        return self._context

    @property
    def controller(self) -> Controller:
        """
        :return: The server controller
        :rtype: trame_server.controller.Controller
        """
        return self._controller

    @property
    def ui(self) -> VirtualNodeManager:
        """
        :return: The server VirtualNode manager
        :rtype: trame_server.ui.VirtualNodeManager
        """
        return self._ui

    @property
    def running(self) -> bool:
        """Return True if the server is currently starting or running."""
        if self.root_server != self:
            return self.root_server.running

        return self._running_stage > 1

    @property
    def network_completion(self):
        """Return a future to await if you want to ensure that any pending network call
        have been issued before locking the server"""
        return asyncio.ensure_future(self.context.network_monitor.completion())

    @property
    def ready(self):
        """Return a future that will resolve once the server is ready"""
        if self.root_server != self:
            return self.root_server.ready

        if self._running_future is None:
            self._running_future = asyncio.get_running_loop().create_future()

        return self._running_future

    def _resolve_ready_future(self, result=None, exception=None) -> None:
        if self.ready.done():
            return

        if exception:
            self.ready.set_exception(exception)
        else:
            self.ready.set_result(result)

    # -------------------------------------------------------------------------
    # API for network handling
    # -------------------------------------------------------------------------

    def get_server_state(self):
        """Return the current server state"""
        return {
            "name": self._name,
            "state": self.state.initial,
        }

    def clear_state_client_cache(self, *state_names):
        protocol = self.protocol
        if protocol:
            protocol.clear_state_client_cache(*state_names)

    # -------------------------------------------------------------------------

    def add_protocol_to_configure(self, configure_protocol_fn):
        """
        Register function that will be called with a wslink.ServerProtocol
        when the server start and is ready for registering new wslink.Protocol.

        :param configure_protocol_fn: A function to be called later with a
                                      wslink.ServerProtocol as argument.
        """
        if self.root_server != self:
            self.root_server.add_protocol_to_configure(configure_protocol_fn)
            return

        self._protocols_to_configure.append(configure_protocol_fn)

    @property
    def protocol(self):
        """Return the server root protocol"""
        if self.root_server != self:
            return self.root_server.protocol

        return self._root_protocol

    # -------------------------------------------------------------------------

    def protocol_call(self, method, *args, **kwargs):
        """
        Call a registered protocol method

        :param method: Method registration name
        :type method: str
        :param *args: Set of args to use for that method call
        :param **kwargs: Set of keyword arguments to use for that method call
        :return: transparently return what the called function returns
        """
        if self.protocol:
            pair = self.protocol.getRPCMethod(method)
            if pair:
                obj, func = pair
                return func(obj, *args, **kwargs)
            return None

        error = "Protocol does not exist yet"
        raise ValueError(error)

    def force_state_push(self, *key_names):
        """
        Should only be needed when client corrupted its data and need the server need to send it again.

        :param *args: Set of key names to be send again to the client.
        """
        self.protocol_call(
            "trame.force.push", *[self._translator.translate_key(k) for k in key_names]
        )

    # -------------------------------------------------------------------------
    # Server handling (start/stop/port)
    # -------------------------------------------------------------------------

    def start(
        self,
        port: int | None = None,
        thread: bool = False,
        open_browser: bool | None = None,
        show_connection_info: bool = True,
        disable_logging: bool = False,
        backend: BackendType | None = None,
        follow_symlinks: bool | None = None,
        exec_mode: ExecModeType = "main",
        timeout: int | None = None,
        host: str | None = None,
        **kwargs,
    ):
        """
        Start the server by listening to the provided port or using the
        `--port, -p` command line argument.
        If the server is already starting or started, any further call will be skipped.

        When the exec_mode="main" or "desktop", the method will be blocking.
        If exec_mode="task", the method will return a scheduled task.
        If exec_mode="coroutine", the method will return a coroutine which
        will need to be scheduled by the user.

        :param port: A port number to listen to. When 0 is provided
                     the system will use a random open port.
        :param thread: If the server run in a thread which means
                       we should disable interruption listeners
        :param open_browser: Should we open the system browser with app url.
                             Using the `--server` command line argument is
                             similar to setting it to False.
        :param show_connection_info: Should we print connection URL at startup?
        :param disable_logging: Ask wslink to disable logging
        :param backend: aiohttp by default but could be generic or tornado.
                        This can also be set with the environment variable ``TRAME_BACKEND``.
                        Defaults to ``'aiohttp'``.
        :param exec_mode: main/desktop/task/coroutine
                          specify how the start function should work
        :param timeout: How much second should we wait before automatically
                        stopping the server when no client is connected.
                        Setting it to 0 will disable such auto-shutdown.
        :param host: The hostname used to bind the server. This can also be
                     set with the environment variable ``TRAME_DEFAULT_HOST``.
                     Defaults to ``'localhost'``.
        :param **kwargs: Keyword arguments for capturing optional parameters
                         for wslink server and/or desktop browser
        """
        if self.root_server != self:
            self.root_server.start(
                port=port,
                thread=thread,
                open_browser=open_browser,
                show_connection_info=show_connection_info,
                disable_logging=disable_logging,
                backend=backend,
                exec_mode=exec_mode,
                timeout=timeout,
                host=host,
                **kwargs,
            )
            return None

        if self._running_stage:
            return None

        # Try to bind client if none were added
        if self._www is None:
            from trame_client import module  # noqa: PLC0415

            self.enable_module(module)

        # Apply any header change needed
        self._http_header.apply()

        # Trigger on_server_start life cycle callback
        if self.controller.on_server_start.exists():
            self.controller.on_server_start(self)

        CoreServer.bind_server(self)
        options = self.cli.parse_known_args()[0]

        if backend is None:
            backend = os.environ.get("TRAME_BACKEND", "aiohttp")

        if follow_symlinks is not None:
            options.static_follow_symlinks = follow_symlinks

        if open_browser is None:
            open_browser = not os.environ.get("TRAME_SERVER")

        if options.host == "localhost":
            if host is None:
                host = os.environ.get("TRAME_DEFAULT_HOST", "localhost")
            options.host = host

        if timeout is not None:
            options.timeout = timeout

        if port is not None:
            options.port = port

        if not options.content:
            options.content = self._www

        if thread:
            options.nosignalhandlers = True

        if options.banner:
            from .utils.banner import print_banner  # noqa: PLC0415

            self.controller.on_server_ready.add(print_banner)

        if options.app:
            exec_mode = "desktop"

        if exec_mode == "desktop":
            from .utils.desktop import start_browser  # noqa: PLC0415

            options.port = 0
            exec_mode, show_connection_info, open_browser = "main", False, False
            self.controller.on_server_ready.add(
                lambda **_: start_browser(self, **kwargs)
            )

        # Allow for older wslink versions where this was not an attribute
        reverse_url = getattr(options, "reverse_url", None)

        if not reverse_url and show_connection_info and exec_mode != "task":
            from .utils.server import print_informations  # noqa: PLC0415

            self.controller.on_server_ready.add(lambda **_: print_informations(self))

        if (
            not reverse_url
            and open_browser
            and exec_mode != "task"
            and not options.server
        ):
            from .utils.browser import open_browser  # noqa: PLC0415

            self.controller.on_server_ready.add(lambda **_: open_browser(self))

        if len(self.serve):
            endpoints = []
            for key in self.serve:
                value = self.serve[key]
                if isinstance(value, (list, tuple)):
                    # tuple are use to describe sync loading (affect client)
                    endpoints.append(f"{key}={value[0]}")
                else:
                    endpoints.append(f"{key}={value}")
            options.fsEndpoints = "|".join(endpoints)

        # Reset http delivery
        if options.no_http:
            options.content = ""
            options.fsEndpoints = ""

        self._server_options = options
        CoreServer.configure(options)

        self._running_stage = 1
        task = CoreServer.server_start(
            options,
            **{  # Do a proper merging/override
                **kwargs,
                "disableLogging": disable_logging,
                "backend": backend,
                "exec_mode": exec_mode,
            },
        )

        # Manage exit life cycle unless coroutine
        if exec_mode == "main":
            self._running_stage = 0
            if self.controller.on_server_exited.exists():
                loop = asyncio.get_event_loop()
                for exit_task in self.controller.on_server_exited(
                    **self.state.to_dict()
                ):
                    if inspect.isawaitable(exit_task):
                        loop.run_until_complete(exit_task)
                    elif callable(exit_task):
                        result = exit_task()
                        if inspect.isawaitable(result):
                            loop.run_until_complete(result)
        elif hasattr(task, "add_done_callback"):

            def on_done(task: asyncio.Task) -> None:
                try:
                    task.result()
                    self._running_stage = 0
                    if self.controller.on_server_exited.exists():
                        self.controller.on_server_exited(**self.state.to_dict())
                except asyncio.CancelledError:
                    self._resolve_ready_future(result=False)
                except Exception as e:  # pylint: disable=broad-except
                    logging.exception("Exception raised by task = %r", task)
                    self._resolve_ready_future(exception=e)

            task.add_done_callback(on_done)

        return task

    async def stop(self) -> None:
        """Coroutine for stopping the server"""
        if self.root_server != self:
            await self.root_server.stop()
        elif self._running_stage:
            await self._server.stop()
            self._running_future = None
        self._running_stage = 0

    @property
    def port(self) -> int:
        """Once started, you can retrieve the port used"""
        if self.root_server != self:
            return self.root_server.port

        return self._running_port

    @property
    def server_options(self):
        """Once started, you can retrieve the server options used"""
        return self._server_options



from trame.app import TrameApp
from trame.decorators import controller, trigger
from trame.ui.router import RouterViewLayout
from trame.ui.vuetify3 import SinglePageLayout, VAppLayout
from trame.widgets import router
from trame.widgets import vuetify3 as v3

import trame_server


class FirstApp(TrameApp):
    def __init__(
        self,
        server: trame_server.Server | str | None = None,
        template_name: str = "main",
    ) -> None:
        super().__init__(server)

        self.state.test = "first"

        self._build_ui(template_name)

    @trigger("test")
    def test_trigger(self) -> None:
        print(self.state.test)

    @controller.add("test_controller")
    def test_controller(self) -> None:
        print(self.state.test)

    def _build_ui(self, template_name) -> None:
        with VAppLayout(
            self.server, template_name=template_name, full_height=True
        ), v3.VContainer(), v3.VCard(title="This is the first app"):
            v3.VBtn("Test Trigger", click="console.log(test); trame.trigger('test');")
            v3.VBtn("Test Controller", click=self.ctrl.test_controller)


class SecondApp(TrameApp):
    def __init__(
        self,
        server: trame_server.Server | str | None = None,
        template_name: str = "main",
    ) -> None:
        super().__init__(server)

        self.state.test = "second"

        self._build_ui(template_name)

    @trigger("test")
    def trigger_test(self) -> None:
        print(self.state.test)

    @controller.add("test_controller")
    def test_controller(self) -> None:
        print(self.state.test)

    def _build_ui(self, template_name) -> None:
        with VAppLayout(
            self.server, template_name=template_name, full_height=True
        ), v3.VContainer(), v3.VCard(title="This is the second app"):
            v3.VBtn("Test Trigger", click="console.log(test); trame.trigger('test');")
            v3.VBtn("Test Controller", click=self.ctrl.test_controller)


class MainApp(TrameApp):
    def __init__(self, server: trame_server.Server | str | None = None):
        super().__init__(server)

        self.state.test = "main"

        self._build_ui()

    @trigger("test")
    def test_trigger(self) -> None:
        print(self.state.test)

    def _build_ui(self) -> None:
        # Register routes
        first_layout = RouterViewLayout(self.server, "/")
        second_layout = RouterViewLayout(self.server, "/second")

        FirstApp(
            self.server.create_child_server(prefix="first_route_"),
            template_name=first_layout.template_name,
        )
        SecondApp(
            self.server.create_child_server(prefix="second_route_"),
            template_name=second_layout.template_name,
        )

        with SinglePageLayout(self.server, full_height=True) as layout:
            with layout.toolbar:
                with v3.VStepper(
                    alt_labels=True,
                    editable=True,
                ):
                    v3.VStepperItem(
                        title="First app",
                        value="1",
                        click="$router.push('/')",
                    )
                    v3.VStepperItem(
                        title="Second app",
                        value="2",
                        click="$router.push('/second')",
                    )
                v3.VSpacer()
                v3.VBtn("Test", click="console.log(test); trame.trigger('test')")

            with layout.content:
                router.RouterView()


def main() -> None:
    app = MainApp()
    app.server.start()


if __name__ == "__main__":
    main()



import asyncio
import logging
import weakref

import pytest

from trame_server.controller import FunctionNotImplementedError

logger = logging.getLogger(__name__)


def test_define_later(controller):
    f = controller.func
    with pytest.raises(FunctionNotImplementedError):
        f()

    controller.func = lambda: 3

    assert f() == 3


def test_trigger_name(controller):
    def fn_1(x):
        return x * 2

    def fn_2(x):
        return x * 3

    a_name = controller.trigger_name(fn_1)
    b_name = controller.trigger_name(fn_2)
    a_name_next = controller.trigger_name(fn_1)

    fn_1_r = controller.trigger_fn(a_name)
    fn_2_r = controller.trigger_fn(b_name)

    assert a_name != b_name
    assert a_name == a_name_next
    assert a_name == "trigger__1"
    assert fn_1 is fn_1_r
    assert fn_2 is fn_2_r


def test_composition(controller):
    def fn():
        return 1

    @controller.add("func_attr")
    def fn_1():
        return 1.5

    assert controller.func_attr() == [1.5]

    @controller.add("func_attr", clear=True)
    def fn_2():
        return 2

    @controller.once("func_attr")
    def fn_3():
        return 3

    # get
    f_attr = controller.func
    f_item = controller["func"]

    assert f_attr is f_item

    # set
    controller.func_attr = fn
    controller["func_item"] = fn

    assert controller.func_attr() == [1, 2, 3]
    assert controller.func_attr() == [1, 2]
    assert controller.func_attr() == [1, 2]

    # invalid set
    with pytest.raises(NameError):
        controller.trigger = fn


def test_weakrefs(controller):
    class Obj:
        method_call_count = 0
        destructor_call_count = 0

        def __del__(self):
            Obj.destructor_call_count += 1

        def fn(self):
            Obj.method_call_count += 1
            print("Obj.fn called")
            return 1

    o = Obj()

    controller.func.add(weakref.WeakMethod(o.fn))

    @controller.add("func")
    def fn_1():
        return 1.5

    controller.func()
    assert Obj.method_call_count == 1

    del o

    assert Obj.destructor_call_count == 1

    controller.func()
    assert Obj.method_call_count == 1


@pytest.mark.asyncio
async def test_tasks(controller):
    @controller.add("async_fn")
    def sync_fn_add():
        return 1

    @controller.add_task("async_fn", clear=True)
    async def async_fn():
        await asyncio.sleep(0.01)
        return 2

    @controller.add("async_fn")
    def sync_fn_add_2():
        return 4

    @controller.set("async_fn")
    def set_fn():
        return 5

    result = controller.async_fn()
    assert len(result) == 3
    assert result[0] == 5
    assert result[1] == 4
    assert await result[2] == 2

    result = controller.async_fn()
    assert len(result) == 3

    with pytest.raises(KeyError):
        controller.async_fn.remove(async_fn)
    controller.async_fn.remove_task(async_fn)

    result = controller.async_fn()
    assert len(result) == 2

    controller.async_fn.discard(async_fn)  # no error if missing
    controller.async_fn.discard(sync_fn_add_2)

    assert controller.async_fn() == 5

    @controller.set("async_fn", clear=True)
    def set_fn_2():
        return 10

    assert controller.async_fn() == 10
    assert controller.async_fn.exists()
    controller.async_fn.clear()
    assert not controller.async_fn.exists()


def test_child_controller(controller, server):
    child_server = server.create_child_server(prefix="child_")
    child_controller = child_server.controller

    @child_controller.add("func")
    def fn_1():
        return 1

    assert child_controller.func() == [1]

    @child_controller.add("func", clear=True)
    def fn_2():
        return 2

    @child_controller.once("func")
    def once_fn():
        return 2.5

    @child_controller.set("func")
    def fn_3():
        return 3

    assert child_controller.func() == [3, 2, 2.5]
    assert child_controller.func() == [3, 2]
    assert child_controller.func() == controller.child_func()
